# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""move_deployment_to_env_2

Migration 7287df262dbc did not handle the case of an old deployment format
which led to failures. Migration `dc0fe6de6786` ports the old format the new
one and we can perform the migration to the env again.

Revision ID: bc908ac9a1fc
Revises: dc0fe6de6786
Create Date: 2018-02-22 21:37:21.258560

"""

import copy
import datetime as dt
import uuid

from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection

from rally.common.db import sa_types
from rally import exceptions

# revision identifiers, used by Alembic.
revision = "bc908ac9a1fc"
down_revision = "dc0fe6de6786"
branch_labels = None
depends_on = None


STATUS_MAP = {
    "deploy->init": "INITIALIZING",
    "deploy->started": "INITIALIZING",
    "deploy->finished": "READY",
    "deploy->failed": "FAILED TO CREATE",
    "deploy->inconsistent": "FAILED TO CREATE",
    "deploy->subdeploy": "INITIALIZING",
    "cleanup->started": "CLEANING",
    "cleanup->failed": "READY",
    "cleanup->finished": "READY"
}


deployments_helper = sa.Table(
    "deployments",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("uuid", sa.String(36), nullable=False),
    sa.Column("name", sa.String(255)),
    sa.Column("config", sa_types.MutableJSONEncodedDict()),
    sa.Column("credentials", sa_types.MutableJSONEncodedDict()),
    sa.Column("enum_deployments_status", sa.Enum(*STATUS_MAP.keys())),
    sa.Column("created_at", sa.DateTime),
)

envs_helper = sa.Table(
    "envs",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("uuid", sa.String(36), nullable=False),

    sa.Column("name", sa.String(255)),
    sa.Column("description", sa.Text),
    sa.Column("status", sa.String(36)),

    sa.Column("extras", sa_types.MutableJSONEncodedDict),
    sa.Column("spec", sa_types.MutableJSONEncodedDict),

    sa.Column("created_at", sa.DateTime),
    sa.Column("updated_at", sa.DateTime)
)

platforms_helper = sa.Table(
    "platforms",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("uuid", sa.String(36), nullable=False),
    sa.Column("env_uuid", sa.String(36)),

    sa.Column("status", sa.String(36)),

    sa.Column("plugin_name", sa.String(36)),
    sa.Column("plugin_spec", sa_types.MutableJSONEncodedDict),
    sa.Column("plugin_data", sa_types.MutableJSONEncodedDict),

    sa.Column("platform_name", sa.String(36)),
    sa.Column("platform_data", sa_types.MutableJSONEncodedDict),

    sa.Column("created_at", sa.DateTime),
    sa.Column("updated_at", sa.DateTime)
)

tasks_helper = sa.Table(
    "tasks",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("uuid", sa.String(36), nullable=False),

    sa.Column("env_uuid", sa.String(36)),
    sa.Column("deployment_uuid", sa.String(36))
)

verifications_helper = sa.Table(
    "verifications",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("uuid", sa.String(36), nullable=False),

    sa.Column("env_uuid", sa.String(36)),
    sa.Column("deployment_uuid", sa.String(36))
)


def upgrade():
    connection = op.get_bind()
    inspector = reflection.Inspector.from_engine(connection)
    if "deployments" not in inspector.get_table_names():
        # 7287df262dbc did not fail. nothing to do
        return

    envs = [env["uuid"] for env in connection.execute(envs_helper.select())]

    for deployment in connection.execute(deployments_helper.select()):
        if deployment["uuid"] in envs:
            # this deployment had been migrated by 7287df262dbc. Nothing to do
            continue
        status = "FAILED TO CREATE"
        spec = deployment.config
        extras = {}
        platform_data = None
        if isinstance(spec, dict) and (
                # existing cloud is only one deployment engine which we
                #   continue supporting
                spec.get("type", "") == "ExistingCloud"
                # We know only about one credential type and it doesn't require
                #   writing additional plugins at the moment.
                and (set(spec["creds"]) == {"openstack"}
                     or not spec["creds"])):

            status = STATUS_MAP[deployment.enum_deployments_status]
            extras = deployment.config.get("extra", {})
            if "openstack" in spec["creds"]:
                spec = {"existing@openstack": spec["creds"]["openstack"]}
                creds = copy.deepcopy(spec["existing@openstack"])

                platform_data = {
                    "admin": creds.pop("admin", {}),
                    "users": creds.pop("users", [])
                }
                platform_data["admin"].update(creds)
                for user in platform_data["users"]:
                    user.update(creds)
            else:
                # empty deployment
                spec = {}

        connection.execute(
            envs_helper.insert(),
            [{
                "uuid": deployment.uuid,
                "name": deployment.name,
                "description": "",
                "status": status,
                "spec": spec,
                "extras": extras,
                "created_at": deployment.created_at,
                "updated_at": dt.datetime.utcnow()
            }]
        )
        if platform_data:
            connection.execute(
                platforms_helper.insert(),
                [{
                    "uuid": str(uuid.uuid4()),
                    "env_uuid": deployment.uuid,
                    "status": "READY",
                    "plugin_name": "existing@openstack",
                    "plugin_spec": spec["existing@openstack"],
                    "plugin_data": {},
                    "platform_name": "openstack",
                    "platform_data": platform_data,
                    "created_at": dt.datetime.utcnow(),
                    "updated_at": dt.datetime.utcnow()
                }]
            )

    op.add_column(
        "verifications",
        sa.Column("env_uuid", sa.String(36))
    )
    op.add_column(
        "tasks",
        sa.Column("env_uuid", sa.String(36))
    )

    conn = op.get_bind()

    conn.execute(
        tasks_helper.update().values(
            env_uuid=tasks_helper.c.deployment_uuid)
    )
    conn.execute(
        verifications_helper.update().values(
            env_uuid=verifications_helper.c.deployment_uuid)
    )

    with op.batch_alter_table("tasks") as batch_op:
        batch_op.alter_column("env_uuid", nullable=False)
        batch_op.drop_index("task_deployment")
        batch_op.drop_column("deployment_uuid")

    with op.batch_alter_table("verifications") as batch_op:
        batch_op.alter_column("env_uuid", nullable=False)
        batch_op.drop_column("deployment_uuid")

    op.drop_index("resource_deployment_uuid", "resources")
    op.drop_index("resource_provider_name", "resources")
    op.drop_index("resource_type", "resources")
    op.drop_index("resource_provider_name_and_type", "resources")
    op.drop_table("resources")

    op.drop_index("deployment_uuid", "deployments")
    op.drop_index("deployment_parent_uuid", "deployments")
    op.drop_table("deployments")


def downgrade():
    raise exceptions.DowngradeNotSupported()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from rally import exceptions
from rally.plugins.common.contexts import dummy
from tests.unit import test


class DummyContextTestCase(test.TestCase):
    def test_setup(self):
        dummy.DummyContext({"task": "some_task"}).setup()
        config = {"dummy_context": {"fail_setup": True}}
        self.assertRaises(
            exceptions.RallyException,
            dummy.DummyContext({"task": "some_task", "config": config}).setup)

    def test_cleanup(self):
        dummy.DummyContext({"task": "some_task"}).cleanup()
        config = {"dummy_context": {"fail_cleanup": True}}
        self.assertRaises(
            exceptions.RallyException,
            dummy.DummyContext({"task": "some_task",
                                "config": config}).cleanup)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import mock

from rally.plugins.common.exporters import trends
from tests.unit import test

PATH = "rally.plugins.common.exporters.html"


def get_tasks_results():
    task_id = "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8"
    workload = {"uuid": "uuid",
                "name": "CinderVolumes.list_volumes",
                "description": "List all volumes.",
                "created_at": "2017-06-04T05:14:44",
                "updated_at": "2017-06-04T05:15:14",
                "task_uuid": task_id,
                "position": 0,
                "data": {"raw": []},
                "full_duration": 29.969523191452026,
                "load_duration": 2.03029203414917,
                "hooks": [],
                "runner": {},
                "runner_type": "runner_type",
                "args": {},
                "contexts": {},
                "contexts_results": [],
                "min_duration": 0.0,
                "max_duration": 1.0,
                "start_time": 0,
                "statistics": {},
                "failed_iteration_count": 0,
                "total_iteration_count": 10,
                "sla": {},
                "sla_results": {"sla": []},
                "pass_sla": True
                }
    task = {
        "uuid": task_id,
        "title": "task",
        "description": "description",
        "status": "finished",
        "env_uuid": "env-uuid",
        "env_name": "env-name",
        "tags": [],
        "created_at": "2017-06-04T05:14:44",
        "updated_at": "2017-06-04T05:15:14",
        "pass_sla": True,
        "task_duration": 2.0,
        "subtasks": [
            {"uuid": "subtask_uuid",
             "title": "subtask",
             "description": "description",
             "status": "finished",
             "run_in_parallel": True,
             "created_at": "2017-06-04T05:14:44",
             "updated_at": "2017-06-04T05:15:14",
             "sla": {},
             "duration": 29.969523191452026,
             "task_uuid": task_id,
             "workloads": [workload]}
        ]}
    return [task]


class TrendsExporterTestCase(test.TestCase):

    @mock.patch("%s.plot.trends" % PATH, return_value="html")
    def test_generate(self, mock_plot_trends):
        reporter = trends.TrendsExporter("task_results", None)

        self.assertEqual({"print": "html"}, reporter.generate())

        mock_plot_trends.assert_called_once_with(
            "task_results", False)

        reporter = trends.TrendsExporter("task_results",
                                         output_destination="path")
        self.assertEqual({"files": {"path": "html"},
                          "open": "file://" + os.path.abspath("path")},
                         reporter.generate())



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import datetime as dt
import os

import ddt
import mock

from rally.common import utils
from rally.plugins.common.verification import reporters
from tests.unit import test


PATH = "rally.plugins.common.verification.reporters"


def get_verifications():
    tests_1 = {
        "some.test.TestCase.test_foo[id=iiiidddd;smoke]":
            {"name": "some.test.TestCase.test_foo",
             "tags": ["smoke", "id"],
             "status": "success",
             "timestamp": "2001-01-01T00:01:00",
             "duration": "8"},
        "some.test.TestCase.test_skipped":
            {"name": "some.test.TestCase.test_skipped",
             "status": "skip",
             "timestamp": "2001-01-01T00:02:00",
             "reason": "Skipped until Bug: 666 is resolved.",
             "duration": "0"},
        "some.test.TestCase.test_xfail":
            {"name": "some.test.TestCase.test_xfail",
             "status": "xfail",
             "timestamp": "2001-01-01T00:03:00",
             "reason": "something",
             "traceback": "HEEELP",
             "duration": "3"},
        "some.test.TestCase.test_uxsuccess":
            {"name": "some.test.TestCase.test_uxsuccess",
             "status": "uxsuccess",
             "timestamp": "2001-01-01T00:04:00",
             "reason": "It should fail since I expect it",
             "duration": "3"}

    }
    tests_2 = {
        "some.test.TestCase.test_foo[id=iiiidddd;smoke]":
            {"name": "some.test.TestCase.test_foo",
             "tags": ["smoke", "id"],
             "status": "success",
             "timestamp": "2001-02-01T00:01:00",
             "duration": "8"},
        "some.test.TestCase.test_failed":
            {"name": "some.test.TestCase.test_failed",
             "status": "fail",
             "timestamp": "2001-02-01T00:02:00",
             "traceback": "HEEEEEEELP",
             "duration": "8"},
        "some.test.TestCase.test_skipped":
            {"name": "some.test.TestCase.test_skipped",
             "status": "skip",
             "timestamp": "2001-02-01T00:03:00",
             "reason": "Skipped until Bug: 666 is resolved.",
             "duration": "0"},
        "some.test.TestCase.test_xfail":
            {"name": "some.test.TestCase.test_xfail",
             "status": "xfail",
             "timestamp": "2001-02-01T00:04:00",
             "reason": "something",
             "traceback": "HEEELP",
             "duration": "4"}
    }
    tests_3 = {
        "some.test.TestCase.test_foo[id=iiiidddd;smoke]":
            {"name": "some.test.TestCase.test_foo",
             "tags": ["smoke", "id"],
             "status": "success",
             "timestamp": "2001-03-01T00:01:00",
             "duration": "8"},
        "some.test.TestCase.test_failed":
            {"name": "some.test.TestCase.test_failed",
             "status": "fail",
             "timestamp": "2001-03-01T00:02:00",
             "traceback": "HEEEEEEELP",
             "duration": "7"},
        "some.test.TestCase.test_skipped":
            {"name": "some.test.TestCase.test_skipped",
             "status": "skip",
             "timestamp": "2001-03-01T00:03:00",
             "reason": "Skipped until Bug: 666 is resolved.",
             "duration": "0"},
        "some.test.TestCase.test_xfail":
            {"name": "some.test.TestCase.test_xfail",
             "status": "xfail",
             "timestamp": "2001-03-01T00:04:00",
             "reason": "something",
             "traceback": "HEEELP",
             "duration": "3"}
    }

    return [
        utils.Struct(uuid="foo-bar-1",
                     created_at=dt.datetime(2001, 1, 1),
                     updated_at=dt.datetime(2001, 1, 2),
                     status="finished",
                     run_args="set_name=compute",
                     tests_duration=1.111,
                     tests_count=9,
                     skipped=0,
                     success=3,
                     expected_failures=3,
                     unexpected_success=2,
                     failures=1,
                     tests=tests_1),
        utils.Struct(uuid="foo-bar-2",
                     created_at=dt.datetime(2002, 1, 1),
                     updated_at=dt.datetime(2002, 1, 2),
                     status="finished",
                     run_args="set_name=full",
                     tests_duration=22.222,
                     tests_count=99,
                     skipped=0,
                     success=33,
                     expected_failures=33,
                     unexpected_success=22,
                     failures=11,
                     tests=tests_2),
        utils.Struct(uuid="foo-bar-3",
                     created_at=dt.datetime(2003, 1, 1),
                     updated_at=dt.datetime(2003, 1, 2),
                     status="finished",
                     run_args="set_name=full",
                     tests_duration=33.333,
                     tests_count=99,
                     skipped=0,
                     success=33,
                     expected_failures=33,
                     unexpected_success=22,
                     failures=11,
                     tests=tests_3)
    ]


class JSONReporterTestCase(test.TestCase):
    def test_validate(self):
        # nothing should fail
        reporters.JSONReporter.validate(mock.Mock())
        reporters.JSONReporter.validate("")
        reporters.JSONReporter.validate(None)

    def test__generate(self):
        reporter = reporters.JSONReporter(get_verifications(), None)
        report = reporter._generate()

        self.assertEqual(
            collections.OrderedDict(
                [("foo-bar-1", {"status": "finished",
                                "started_at": "2001-01-01T00:00:00",
                                "finished_at": "2001-01-02T00:00:00",
                                "tests_duration": 1.111,
                                "tests_count": 9,
                                "run_args": "set_name=compute",
                                "skipped": 0,
                                "success": 3,
                                "unexpected_success": 2,
                                "expected_failures": 3,
                                "failures": 1}),
                 ("foo-bar-2", {"status": "finished",
                                "started_at": "2002-01-01T00:00:00",
                                "finished_at": "2002-01-02T00:00:00",
                                "tests_duration": 22.222,
                                "tests_count": 99,
                                "run_args": "set_name=full",
                                "skipped": 0,
                                "success": 33,
                                "unexpected_success": 22,
                                "expected_failures": 33,
                                "failures": 11}),
                 ("foo-bar-3", {"status": "finished",
                                "started_at": "2003-01-01T00:00:00",
                                "finished_at": "2003-01-02T00:00:00",
                                "tests_duration": 33.333,
                                "tests_count": 99,
                                "run_args": "set_name=full",
                                "skipped": 0,
                                "success": 33,
                                "unexpected_success": 22,
                                "expected_failures": 33,
                                "failures": 11})]),
            report["verifications"])

        self.assertEqual({
            "some.test.TestCase.test_foo[id=iiiidddd;smoke]": {
                "by_verification": {"foo-bar-1": {"duration": "8",
                                                  "status": "success"},
                                    "foo-bar-2": {"duration": "8",
                                                  "status": "success"},
                                    "foo-bar-3": {"duration": "8",
                                                  "status": "success"}
                                    },
                "name": "some.test.TestCase.test_foo",
                "tags": ["smoke", "id"]},
            "some.test.TestCase.test_failed": {
                "by_verification": {"foo-bar-2": {"details": "HEEEEEEELP",
                                                  "duration": "8",
                                                  "status": "fail"},
                                    "foo-bar-3": {"details": "HEEEEEEELP",
                                                  "duration": "7",
                                                  "status": "fail"}},
                "name": "some.test.TestCase.test_failed",
                "tags": []},
            "some.test.TestCase.test_skipped": {
                "by_verification": {
                    "foo-bar-1": {
                        "details": "Skipped until Bug: https://launchpad.net/"
                                   "bugs/666 is resolved.",
                        "duration": "0",
                        "status": "skip"},
                    "foo-bar-2": {
                        "details": "Skipped until Bug: https://launchpad.net/"
                                   "bugs/666 is resolved.",
                        "duration": "0",
                        "status": "skip"},
                    "foo-bar-3": {
                        "details": "Skipped until Bug: https://launchpad.net/"
                                   "bugs/666 is resolved.",
                        "duration": "0",
                        "status": "skip"}},
                "name": "some.test.TestCase.test_skipped",
                "tags": []},
            "some.test.TestCase.test_xfail": {
                "by_verification": {
                    "foo-bar-1": {"details": "something\n\nHEEELP",
                                  "duration": "3",
                                  "status": "xfail"},
                    "foo-bar-2": {"details": "something\n\nHEEELP",
                                  "duration": "4",
                                  "status": "xfail"},
                    "foo-bar-3": {"details": "something\n\nHEEELP",
                                  "duration": "3",
                                  "status": "xfail"}},
                "name": "some.test.TestCase.test_xfail",
                "tags": []},
            "some.test.TestCase.test_uxsuccess": {
                "name": "some.test.TestCase.test_uxsuccess",
                "tags": [],
                "by_verification": {"foo-bar-1": {
                    "details": "It should fail since I expect it",
                    "duration": "3",
                    "status": "uxsuccess"}}}},
            report["tests"])

    @mock.patch("%s.json.dumps" % PATH)
    @mock.patch("%s.JSONReporter._generate" % PATH)
    def test_generate(self, mock__generate, mock_dumps):
        reporter = reporters.JSONReporter([], output_destination=None)
        self.assertEqual({"print": mock_dumps.return_value},
                         reporter.generate())
        mock__generate.assert_called_once_with()
        mock_dumps.assert_called_once_with(mock__generate.return_value,
                                           indent=4)

        mock__generate.reset_mock()
        mock_dumps.reset_mock()

        path = "some_path"
        reporter = reporters.JSONReporter([], output_destination=path)
        self.assertEqual({"files": {path: mock_dumps.return_value},
                          "open": path}, reporter.generate())
        mock__generate.assert_called_once_with()
        mock_dumps.assert_called_once_with(mock__generate.return_value,
                                           indent=4)


@ddt.ddt
class HTMLReporterTestCase(test.TestCase):
    @mock.patch("%s.ui_utils" % PATH)
    @mock.patch("%s.json.dumps" % PATH)
    @ddt.data((reporters.HTMLReporter, False),
              (reporters.HTMLStaticReporter, True))
    @ddt.unpack
    def test_generate(self, cls, include_libs, mock_dumps, mock_ui_utils):
        mock_render = mock_ui_utils.get_template.return_value.render

        reporter = cls(get_verifications(), None)

        self.assertEqual({"print": mock_render.return_value},
                         reporter.generate())
        mock_render.assert_called_once_with(data=mock_dumps.return_value,
                                            include_libs=include_libs)
        mock_ui_utils.get_template.assert_called_once_with(
            "verification/report.html")

        self.assertEqual(1, mock_dumps.call_count)
        args, kwargs = mock_dumps.call_args
        self.assertTrue(not kwargs)
        self.assertEqual(1, len(args))
        ctx = args[0]
        self.assertEqual({"uuids", "verifications", "tests",
                          "show_comparison_note"},
                         set(ctx.keys()))
        self.assertEqual(["foo-bar-1", "foo-bar-2", "foo-bar-3"],
                         list(ctx["uuids"]))
        self.assertTrue(ctx["show_comparison_note"])
        self.assertEqual({
            "some.test.TestCase.test_foo[id=iiiidddd;smoke]": {
                "by_verification": {"foo-bar-1": {"details": None,
                                                  "duration": "8",
                                                  "status": "success"},
                                    "foo-bar-2": {"details": None,
                                                  "duration": "8 (+0.0)",
                                                  "status": "success"},
                                    "foo-bar-3": {"details": None,
                                                  "duration": "8 (+0.0)",
                                                  "status": "success"}},
                "has_details": False,
                "name": "some.test.TestCase.test_foo",
                "tags": ["smoke", "id"]},
            "some.test.TestCase.test_failed": {
                "by_verification": {"foo-bar-2": {"details": "HEEEEEEELP",
                                                  "duration": "8",
                                                  "status": "fail"},
                                    "foo-bar-3": {"details": "HEEEEEEELP",
                                                  "duration": "7 (-1.0)",
                                                  "status": "fail"}},
                "has_details": True,
                "name": "some.test.TestCase.test_failed",
                "tags": []},
            "some.test.TestCase.test_skipped": {
                "by_verification": {
                    "foo-bar-1": {"details": "Skipped until Bug: https://laun"
                                             "chpad.net/bugs/666 is resolved.",
                                  "duration": "",
                                  "status": "skip"},
                    "foo-bar-2": {"details": "Skipped until Bug: https://laun"
                                             "chpad.net/bugs/666 is resolved.",
                                  "duration": "",
                                  "status": "skip"},
                    "foo-bar-3": {"details": "Skipped until Bug: https://laun"
                                             "chpad.net/bugs/666 is resolved.",
                                  "duration": "",
                                  "status": "skip"}},
                "has_details": True,
                "name": "some.test.TestCase.test_skipped",
                "tags": []},
            "some.test.TestCase.test_xfail": {
                "by_verification": {
                    "foo-bar-1": {"details": "something\n\nHEEELP",
                                  "duration": "3",
                                  "status": "xfail"},
                    "foo-bar-2": {"details": "something\n\nHEEELP",
                                  "duration": "4 (+1.0)",
                                  "status": "xfail"},
                    "foo-bar-3": {"details": "something\n\nHEEELP",
                                  "duration": "3 (+0.0)",
                                  "status": "xfail"}},
                "has_details": True,
                "name": "some.test.TestCase.test_xfail",
                "tags": []},
            "some.test.TestCase.test_uxsuccess": {
                "name": "some.test.TestCase.test_uxsuccess",
                "tags": [],
                "has_details": True,
                "by_verification": {"foo-bar-1": {
                    "details": "It should fail since I expect it",
                    "duration": "3",
                    "status": "uxsuccess"}}}},
            ctx["tests"])

    def test_generate_raw_html_report(self):
        reporter = reporters.HTMLReporter(get_verifications(), None)
        report = reporter.generate()
        self.assertTrue(report["print"])


class JUnitXMLReporterTestCase(test.TestCase):
    @mock.patch("%s.dt" % PATH)
    @mock.patch("%s.version.version_string" % PATH)
    def test_generate(self, mock_version_string, mock_dt):
        mock_dt.datetime.utcnow.return_value.strftime.return_value = "TIME"
        # release when junit reporter was introduced
        mock_version_string.return_value = "0.8.0"
        with open(os.path.join(os.path.dirname(__file__),
                               "junit_report.xml")) as f:
            expected_report = f.read()

        junit_reporter = reporters.JUnitXMLReporter(get_verifications(), None)
        self.assertEqual({"print": expected_report}, junit_reporter.generate())

        dest = "path"
        junit_reporter = reporters.JUnitXMLReporter(get_verifications(), dest)
        self.assertEqual({"open": dest,
                          "files": {dest: expected_report}},
                         junit_reporter.generate())



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools

from tests.functional import utils


class PluginTestCase(testtools.TestCase):

    def test_show_one(self):
        rally = utils.Rally()
        result = rally("plugin show Dummy.dummy")
        self.assertIn("NAME", result)
        self.assertIn("PLATFORM", result)
        self.assertIn("Dummy.dummy", result)
        self.assertIn("MODULE", result)

    def test_show_multiple(self):
        rally = utils.Rally()
        result = self.assertRaises(utils.RallyCliError,
                                   rally, "plugin show Dummy")
        self.assertIn("Multiple plugins found:", result.output)
        self.assertIn("Dummy.dummy", result.output)
        self.assertIn("Dummy.dummy_exception", result.output)
        self.assertIn("Dummy.dummy_random_fail_in_atomic", result.output)

    def test_show_not_found(self):
        rally = utils.Rally()
        name = "Dummy666666"
        result = self.assertRaises(utils.RallyCliError,
                                   rally, "plugin show %s" % name)
        self.assertIn("Plugin %s not found" % name, result.output)

    def test_show_not_found_in_specific_platform(self):
        rally = utils.Rally()
        name = "Dummy"
        platform = "non_existing"
        result = self.assertRaises(
            utils.RallyCliError,
            rally, "plugin show --name %(name)s --platform %(platform)s"
                   % {"name": name, "platform": platform})
        self.assertIn(
            "Plugin %(name)s@%(platform)s not found"
            % {"name": name, "platform": platform},
            result.output)

    def test_list(self):
        rally = utils.Rally()
        result = rally("plugin list Dummy")
        self.assertIn("Dummy.dummy", result)
        self.assertIn("Dummy.dummy_exception", result)
        self.assertIn("Dummy.dummy_random_fail_in_atomic", result)

    def test_list_not_found_platform(self):
        rally = utils.Rally()
        result = rally("plugin list --platform some")
        self.assertIn("Platform some not found", result)

    def test_list_not_found_name(self):
        rally = utils.Rally()
        result = rally("plugin list Dummy2222")
        self.assertIn("Plugin Dummy2222 not found", result)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defines interface for DB access.

The underlying driver is loaded as a :class:`LazyPluggable`.

Functions in this module are imported into the rally.common.db namespace.
Call these functions from rally.common.db namespace, not the
rally.common.db.api namespace.

All functions in this module return objects that implement a dictionary-like
interface. Currently, many of these objects are sqlalchemy objects that
implement a dictionary interface. However, a future goal is to have all of
these objects be simple dictionaries.


**Related Flags**

:backend:  string to lookup in the list of LazyPluggable backends.
           `sqlalchemy` is the only supported backend right now.

:connection:  string specifying the sqlalchemy connection to use, like:
              `sqlite:///var/lib/cinder/cinder.sqlite`.

:enable_new_services:  when adding a new service to the database, is it in the
                       pool of available hardware (Default: True)

"""

import datetime as dt
import functools
import tempfile
import time

from oslo_db import exception as db_exc
from oslo_db import options as db_options
from oslo_db.sqlalchemy import session as db_session
import sqlalchemy as sa
import sqlalchemy.orm   # noqa

from rally.common import cfg
from rally.common.db import models
from rally import consts
from rally import exceptions
from rally.task.processing import charts


CONF = cfg.CONF

db_options.set_defaults(
    CONF, connection="sqlite:///%s/rally.sqlite" % tempfile.gettempdir())

_FACADE = None
_SESSION_MAKER = None


def _create_facade_lazily():
    global _FACADE

    if _FACADE is None:
        _FACADE = db_session.EngineFacade.from_config(CONF)

    return _FACADE


def get_engine():
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session():
    global _SESSION_MAKER

    if not _SESSION_MAKER:
        _SESSION_MAKER = sa.orm.sessionmaker()
        _SESSION_MAKER.configure(bind=get_engine())

    return _SESSION_MAKER()


def engine_reset():
    global _FACADE, _SESSION_MAKER
    _FACADE = None
    _SESSION_MAKER = None


def serialize(data):
    if data is None:
        return None
    if isinstance(data, (int,
                         str,
                         dt.date,
                         dt.time,
                         float,
                         )):
        return data
    if isinstance(data, models.RallyBase):
        result = data.as_dict()
        for k in result:
            result[k] = serialize(result[k])
        return result
    if isinstance(data, (list, tuple)):
        return [serialize(d) for d in data]
    if isinstance(data, dict):
        result = {}
        for k in data:
            result[k] = serialize(data[k])
        return result

    raise ValueError("Failed to serialize %r data type." % type(data).__name__)


def with_session(f):

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        session = get_session()
        session.expire_on_commit = False
        try:
            result = f(session, *args, **kwargs)
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

        return serialize(result)

    return wrapper


@with_session
def tags_get(session, uuid, tag_type):
    query = session.query(models.Tag.tag).filter_by(uuid=uuid, type=tag_type)
    return [t.tag for t in query.distinct().all()]


def _uuids_by_tags_get(session, tag_type, tags):
    tags = (session.query(models.Tag.uuid)
                   .filter(models.Tag.type == tag_type,
                           models.Tag.tag.in_(tags)).distinct())
    return [t.uuid for t in tags.all()]


def _task_workload_data_get_all(session, workload_uuid):
    results = (session.query(models.WorkloadData)
                      .filter_by(workload_uuid=workload_uuid)
                      .order_by(models.WorkloadData.chunk_order.asc()))

    return sorted([raw for workload_data in results
                   for raw in workload_data.chunk_data["raw"]],
                  key=lambda x: x["timestamp"])


def _subtasks_get_all_by_task_uuid(session, task_uuid):
    result = session.query(models.Subtask).filter_by(task_uuid=task_uuid).all()
    subtasks = []
    for subtask in result:
        subtask = subtask.as_dict()
        subtask["workloads"] = []
        workloads = (session.query(models.Workload).
                     filter_by(subtask_uuid=subtask["uuid"]).all())
        for workload in workloads:
            workload = workload.as_dict()
            workload["data"] = _task_workload_data_get_all(
                session, workload["uuid"])
            subtask["workloads"].append(workload)
        subtasks.append(subtask)
    return subtasks


@with_session
def task_get(session, uuid=None, detailed=False):

    task = session.query(models.Task).filter_by(uuid=uuid).first()
    if not task:
        raise exceptions.DBRecordNotFound(
            criteria="uuid: %s" % uuid, table="tasks")
    task = task.as_dict()
    task["tags"] = sorted(tags_get(uuid, consts.TagType.TASK))

    if detailed:
        task["subtasks"] = _subtasks_get_all_by_task_uuid(session, uuid)

    return task


@with_session
def task_get_status(session, uuid=None):
    task = (session.query(models.Task)
                   .options(sa.orm.load_only("status"))
                   .filter_by(uuid=uuid).first())
    if not task:
        raise exceptions.DBRecordNotFound(
            criteria="uuid: %s" % uuid, table="tasks")

    return task.status


@with_session
def task_create(session, values):
    tags = values.pop("tags", [])
    # TODO(ikhudoshyn): currently 'input_task'
    # does not come in 'values'
    # After completely switching to the new
    # DB schema in API we should reconstruct
    # input_task's from associated workloads
    # the same is true for 'pass_sla',
    # 'task_duration', 'validation_result'
    # and 'validation_duration'
    task = models.Task(**values)
    session.add(task)
    session.commit()
    task = task.as_dict()

    if tags:
        session.bulk_save_objects(
            [models.Tag(uuid=task["uuid"], tag=t,
                        type=consts.TagType.TASK)
             for t in set(tags)])
    task["tags"] = tags
    return task


@with_session
def task_update(session, uuid, values):
    values.pop("uuid", None)
    tags = values.pop("tags", None)

    task = session.query(models.Task).filter_by(uuid=uuid).first()
    if not task:
        raise exceptions.DBRecordNotFound(
            criteria="uuid: %s" % uuid, table="tasks")
    task.update(values)
    task = task.as_dict()

    if tags is not None:
        # TODO(boris-42): create separate method for tags editing
        tags_in_db = session.query(models.Tag.tag).filter_by(
            uuid=uuid, type=consts.TagType.TASK).distinct()
        new_tags = set(tags) - set(tags_in_db)
        removed_tags = set(tags_in_db) - set(tags)

        (session.query(models.Tag)
                .filter_by(uuid=uuid, type=consts.TagType.TASK)
                .filter(models.Tag.tag.in_(removed_tags))
                .delete(synchronize_session=False))

        if new_tags:
            session.bulk_save_objects(
                [models.Tag(uuid=uuid, tag=t, type=consts.TagType.TASK)
                 for t in set(new_tags)])
        task["tags"] = tags
    else:
        task["tags"] = []
    return task


@with_session
def task_update_status(session, uuid, status, allowed_statuses):
    result = (session.query(models.Task)
                     .filter(models.Task.uuid == uuid,
                             models.Task.status.in_(allowed_statuses))
                     .update({"status": status}, synchronize_session=False))
    if not result:
        raise exceptions.DBRecordNotFound(
            criteria="uuid=%(uuid)s and status in [%(statuses)s]"
                     % {"uuid": uuid, "statuses": ", ".join(allowed_statuses)},
            table="tasks")
    return result


@with_session
def task_list(session, status=None, env=None, tags=None, uuids_only=False):
    tasks = []
    query = session.query(models.Task)

    filters = {}
    if status is not None:
        filters["status"] = status
    if env is not None:
        filters["env_uuid"] = env_get(env)["uuid"]
    if filters:
        query = query.filter_by(**filters)

    if tags:
        uuids = _uuids_by_tags_get(session, consts.TagType.TASK, tags)
        if not uuids:
            return []
        query = query.filter(models.Task.uuid.in_(uuids))

    if uuids_only:
        query = query.options(sa.orm.load_only("uuid"))

    for task in query.all():
        task = task.as_dict()
        if not uuids_only:
            task["tags"] = sorted(tags_get(task["uuid"], consts.TagType.TASK))
        tasks.append(task)

    return tasks


@with_session
def task_delete(session, uuid, status=None):
    (session.query(models.WorkloadData).filter_by(task_uuid=uuid).
     delete(synchronize_session=False))

    (session.query(models.Workload).filter_by(task_uuid=uuid).
     delete(synchronize_session=False))

    (session.query(models.Subtask).filter_by(task_uuid=uuid).
     delete(synchronize_session=False))

    (session.query(models.Tag).filter_by(
        uuid=uuid, type=consts.TagType.TASK).
     delete(synchronize_session=False))

    query = session.query(models.Task).filter_by(uuid=uuid)
    if status:
        count = query.filter_by(status=status).delete(
            synchronize_session="fetch")
    else:
        count = query.delete(synchronize_session="fetch")
    if not count:
        if status is not None:
            task = query.first()
            if task:
                raise exceptions.DBConflict(
                    "Task `%(uuid)s` in `%(actual)s` status but "
                    "`%(require)s` is required."
                    % {"uuid": uuid,
                       "require": status, "actual": task.status})

        raise exceptions.DBRecordNotFound(
            criteria="uuid: %s" % uuid, table="tasks")


@with_session
def subtask_create(session, task_uuid, title, description=None, contexts=None):
    subtask = models.Subtask(task_uuid=task_uuid,
                             title=title,
                             description=description or "",
                             contexts=contexts or {})
    session.add(subtask)
    return subtask


@with_session
def subtask_update(session, subtask_uuid, values):
    subtask = session.query(models.Subtask).filter_by(
        uuid=subtask_uuid).first()
    subtask.update(values)
    return subtask


@with_session
def workload_get(session, workload_uuid):
    return session.query(models.Workload).filter_by(uuid=workload_uuid).first()


@with_session
def workload_create(session, task_uuid, subtask_uuid, name, description,
                    position, runner, runner_type, hooks, contexts, sla, args):
    workload = models.Workload(task_uuid=task_uuid,
                               subtask_uuid=subtask_uuid,
                               name=name,
                               description=description,
                               position=position,
                               runner=runner,
                               runner_type=runner_type,
                               hooks=hooks,
                               contexts=contexts or {},
                               sla=sla,
                               args=args)
    session.add(workload)
    return workload


@with_session
def workload_data_create(session, task_uuid, workload_uuid, chunk_order, data):
    workload_data = models.WorkloadData(task_uuid=task_uuid,
                                        workload_uuid=workload_uuid)

    raw_data = data.get("raw", [])
    iter_count = len(raw_data)

    failed_iter_count = 0

    started_at = float("inf")
    finished_at = 0
    for d in raw_data:
        if d.get("error"):
            failed_iter_count += 1

        timestamp = d["timestamp"]
        duration = d["duration"]
        finished = timestamp + duration

        if timestamp < started_at:
            started_at = timestamp

        if finished > finished_at:
            finished_at = finished

    now = time.time()
    if started_at == float("inf"):
        started_at = now
    if finished_at == 0:
        finished_at = now

    workload_data.update({
        "task_uuid": task_uuid,
        "workload_uuid": workload_uuid,
        "chunk_order": chunk_order,
        "iteration_count": iter_count,
        "failed_iteration_count": failed_iter_count,
        "chunk_data": {"raw": raw_data},
        # TODO(ikhudoshyn)
        "chunk_size": 0,
        "compressed_chunk_size": 0,
        "started_at": dt.datetime.fromtimestamp(started_at),
        "finished_at": dt.datetime.fromtimestamp(finished_at)
    })
    session.add(workload_data)
    return workload_data


@with_session
def workload_set_results(session, workload_uuid, subtask_uuid, task_uuid,
                         load_duration, full_duration, start_time,
                         sla_results, contexts_results, hooks_results=None):
    workload_results = _task_workload_data_get_all(session, workload_uuid)

    iter_count = len(workload_results)

    failed_iter_count = 0
    max_duration = None
    min_duration = None

    for d in workload_results:
        if d.get("error"):
            failed_iter_count += 1

        duration = d.get("duration", 0)

        if max_duration is None or duration > max_duration:
            max_duration = duration

        if min_duration is None or min_duration > duration:
            min_duration = duration

    durations_stat = charts.MainStatsTable(
        {"total_iteration_count": iter_count})

    for itr in workload_results:
        durations_stat.add_iteration(itr)

    sla = sla_results or []
    # NOTE(ikhudoshyn): we call it 'pass_sla'
    # for the sake of consistency with other models
    # so if no SLAs were specified, then we assume pass_sla == True
    success = all([s.get("success") for s in sla])

    session.query(models.Workload).filter_by(
        uuid=workload_uuid).update(
        {
            "sla_results": {"sla": sla},
            "contexts_results": contexts_results,
            "hooks": hooks_results or [],
            "load_duration": load_duration,
            "full_duration": full_duration,
            "min_duration": min_duration,
            "max_duration": max_duration,
            "total_iteration_count": iter_count,
            "failed_iteration_count": failed_iter_count,
            "start_time": start_time,
            "statistics": {"durations": durations_stat.to_dict()},
            "pass_sla": success}
    )
    task_values = {
        "task_duration": models.Task.task_duration + load_duration}
    if not success:
        task_values["pass_sla"] = False

    subtask_values = {
        "duration": models.Subtask.duration + load_duration}
    if not success:
        subtask_values["pass_sla"] = False
    session.query(models.Task).filter_by(uuid=task_uuid).update(
        task_values)

    session.query(models.Subtask).filter_by(uuid=subtask_uuid).update(
        subtask_values)


@with_session
def env_get(session, uuid_or_name):
    env = (session.query(models.Env)
                  .filter(sa.or_(models.Env.uuid == uuid_or_name,
                                 models.Env.name == uuid_or_name))
                  .first())
    if not env:
        raise exceptions.DBRecordNotFound(
            criteria="uuid or name is %s" % uuid_or_name, table="envs")
    return env


@with_session
def env_get_status(session, uuid):
    resp = (session.query(models.Env)
                   .filter_by(uuid=uuid)
                   .options(sa.orm.load_only("status"))
                   .first())
    if not resp:
        raise exceptions.DBRecordNotFound(
            criteria="uuid: %s" % uuid, table="envs")
    return resp.status


@with_session
def env_list(session, status=None):
    query = session.query(models.Env)
    if status:
        query = query.filter_by(status=status)
    return query.all()


@with_session
def env_create(session, name, status, description, extras, config,
               spec, platforms):
    try:
        env_uuid = models.UUID()
        for p in platforms:
            p["env_uuid"] = env_uuid

        env = models.Env(
            name=name, uuid=env_uuid,
            status=status, description=description,
            extras=extras, config=config, spec=spec
        )
        session.add(env)
        session.commit()
        session.bulk_save_objects(
            [models.Platform(**p) for p in platforms])
    except db_exc.DBDuplicateEntry:
        raise exceptions.DBRecordExists(
            field="name", value=name, table="envs")

    return env


@with_session
def env_rename(session, uuid, old_name, new_name):
    try:
        return bool(session.query(models.Env)
                           .filter_by(uuid=uuid, name=old_name)
                           .update({"name": new_name}))
    except db_exc.DBDuplicateEntry:
        raise exceptions.DBRecordExists(
            field="name", value=new_name, table="envs")


@with_session
def env_update(session, uuid, description=None, extras=None, config=None):
    values = {}
    if description is not None:
        values["description"] = description
    if extras is not None:
        values["extras"] = extras
    if config is not None:
        values["config"] = config

    if not values:
        return True

    return bool(session.query(models.Env).filter_by(uuid=uuid).update(values))


@with_session
def env_set_status(session, uuid, old_status, new_status):
    count = (session.query(models.Env)
                    .filter_by(uuid=uuid, status=old_status)
                    .update({"status": new_status}))
    if count:
        return True
    raise exceptions.DBConflict("Env %s should be in status %s actual %s"
                                % (uuid, old_status, env_get_status(uuid)))


@with_session
def env_delete_cascade(session, uuid):
    for model in [models.Task, models.Verification, models.Platform]:
        session.query(model).filter_by(env_uuid=uuid).delete()
    session.query(models.Env).filter_by(uuid=uuid).delete()


@with_session
def platforms_list(session, env_uuid):
    return session.query(models.Platform).filter_by(env_uuid=env_uuid).all()


@with_session
def platform_get(session, uuid):
    p = session.query(models.Platform).filter_by(uuid=uuid).first()
    if not p:
        raise exceptions.DBRecordNotFound(
            criteria="uuid = %s" % uuid, table="platforms")
    return p


@with_session
def platform_set_status(session, uuid, old_status, new_status):
    count = (session.query(models.Platform)
                    .filter_by(uuid=uuid, status=old_status)
                    .update({"status": new_status}))
    if count:
        return True

    platform = platform_get(uuid)
    raise exceptions.DBConflict(
        "Platform %s should be in status %s actual %s"
        % (uuid, old_status, platform["status"]))


@with_session
def platform_set_data(session, uuid, platform_data=None, plugin_data=None):
    values = {}
    if platform_data is not None:
        values["platform_data"] = platform_data
    if plugin_data is not None:
        values["plugin_data"] = plugin_data

    if not values:
        return True

    return bool(
        session.query(models.Platform).filter_by(uuid=uuid).update(values))


@with_session
def verifier_create(session, name, vtype, platform, source, version,
                    system_wide, extra_settings=None):
    verifier = models.Verifier(name=name, type=vtype, platform=platform,
                               source=source, extra_settings=extra_settings,
                               version=version, system_wide=system_wide)
    session.add(verifier)
    return verifier


@with_session
def verifier_get(session, verifier_id):
    return _verifier_get(session, verifier_id)


def _verifier_get(session, verifier_id):
    verifier = (session.query(models.Verifier)
                       .filter(sa.or_(models.Verifier.name == verifier_id,
                                      models.Verifier.uuid == verifier_id))
                       .first())
    if not verifier:
        raise exceptions.DBRecordNotFound(
            criteria="name or uuid is %s" % verifier_id, table="verifiers")
    return verifier


@with_session
def verifier_list(session, status=None):
    query = session.query(models.Verifier)
    if status:
        query = query.filter_by(status=status)
    return query.all()


@with_session
def verifier_delete(session, verifier_id):
    count = (session.query(models.Verifier)
                    .filter(sa.or_(models.Verifier.name == verifier_id,
                                   models.Verifier.uuid == verifier_id))
                    .delete(synchronize_session=False))
    if not count:
        raise exceptions.DBRecordNotFound(
            criteria="name or uuid is %s" % verifier_id, table="verifiers")


@with_session
def verifier_update(session, verifier_id, **properties):
    verifier = _verifier_get(session, verifier_id)
    verifier.update(properties)
    return verifier


@with_session
def verification_create(session, verifier_id, env, tags=None, run_args=None):
    verifier = _verifier_get(session, verifier_id)
    env = env_get(env)
    verification = models.Verification(verifier_uuid=verifier.uuid,
                                       env_uuid=env["uuid"],
                                       run_args=run_args)
    session.add(verification)
    session.commit()

    if tags:
        session.bulk_save_objects(
            [models.Tag(uuid=verification.uuid, tag=t,
                        type=consts.TagType.VERIFICATION)
             for t in set(tags)]
        )
    return verification


@with_session
def verification_get(session, verification_uuid):
    verification = _verification_get(session, verification_uuid)
    verification.tags = sorted(tags_get(verification.uuid,
                                        consts.TagType.VERIFICATION))
    return verification


def _verification_get(session, verification_uuid):
    verification = session.query(models.Verification).filter_by(
        uuid=verification_uuid).first()
    if not verification:
        raise exceptions.DBRecordNotFound(
            criteria="uuid: %s" % verification_uuid, table="verifications")
    return verification


@with_session
def verification_list(session,
                      verifier_id=None, env=None, tags=None, status=None):
    filter_by = {}
    if verifier_id:
        verifier = _verifier_get(session, verifier_id)
        filter_by["verifier_uuid"] = verifier.uuid
    if env:
        env = env_get(env)
        filter_by["env_uuid"] = env["uuid"]
    if status:
        filter_by["status"] = status

    query = session.query(models.Verification)
    if filter_by:
        query = query.filter_by(**filter_by)

    if tags:
        uuids = _uuids_by_tags_get(session,
                                   consts.TagType.VERIFICATION, tags)
        query = query.filter(models.Verification.uuid.in_(uuids))

    verifications = [verification.as_dict() for verification in query.all()]
    for verification in verifications:
        verification["tags"] = sorted(tags_get(verification["uuid"],
                                               consts.TagType.VERIFICATION))
    return verifications


@with_session
def verification_delete(session, uuid):
    count = session.query(models.Verification).filter_by(uuid=uuid).delete()
    if not count:
        raise exceptions.DBRecordNotFound(criteria="uuid: %s" % uuid,
                                          table="verifications")


@with_session
def verification_update(session, verification_uuid, **properties):
    verification = _verification_get(session, verification_uuid)
    verification.update(properties)
    return verification



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from rally.plugins.common.scenarios.requests import utils
from tests.unit import test


class RequestsTestCase(test.TestCase):

    @mock.patch("requests.request")
    def test__check_request(self, mock_request):
        mock_request.return_value = mock.MagicMock(status_code=200)
        scenario = utils.RequestScenario(test.get_test_context())
        scenario._check_request(status_code=200, url="sample", method="GET")

        self._test_atomic_action_timer(scenario.atomic_actions(),
                                       "requests.check_request")
        mock_request.assert_called_once_with("GET", "sample")

    @mock.patch("requests.request")
    def test_check_wrong_request(self, mock_request):
        mock_request.return_value = mock.MagicMock(status_code=200)
        scenario = utils.RequestScenario(test.get_test_context())

        self.assertRaises(ValueError, scenario._check_request,
                          status_code=201, url="sample", method="GET")



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import json
import os
import re
import sys
import time

import jinja2
import jinja2.meta
import requests
from requests.packages import urllib3

from rally.common import cfg
from rally.common import logging
from rally.common import objects
from rally.common import opts
from rally.common.plugin import discover
from rally.common import utils
from rally.common import version as rally_version
from rally import consts
from rally import exceptions
from rally.task import engine
from rally.task import exporter as texporter
from rally.task import task_cfg
from rally.verification import context as vcontext
from rally.verification import manager as vmanager
from rally.verification import reporter as vreporter


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
API_REQUEST_PREFIX = "/api"


opts.register()


class APIGroup(object):
    def __init__(self, api):
        """Initialize API group.

        :param api: an instance of rally.api.API object
        """
        self.api = api


class _Deployment(APIGroup):

    def _create(self, config, name):
        """Create a deployment.

        :param config: a dict with deployment configuration
        :param name: a str represents a name of the deployment
        :returns: Deployment object
        """

        # NOTE(andreykurilin): the following transformation is a preparatory
        #   step for further refactoring (it will be done soon).
        print_warning = False

        extras = {}
        if "type" in config:
            if config["type"] != "ExistingCloud":
                raise exceptions.RallyException(
                    "You are using deployment type which doesn't exist. Please"
                    " check the latest documentation and fix deployment "
                    "config.")

            config = config["creds"]
            extras = config.get("extra", {})
            print_warning = True

        try:
            deployment = objects.Deployment(name=name, config=config,
                                            extras=extras)
        except exceptions.DBRecordExists:
            if logging.is_debug():
                LOG.exception("Deployment with such name exists")
            raise

        if print_warning:
            new_conf = deployment.env_obj.spec
            LOG.warning(
                "The used config schema is deprecated since Rally 0.10.0. "
                "The new one is much simpler, try it now:\n%s"
                % json.dumps(new_conf, indent=4)
            )

        return deployment

    def create(self, config, name):
        return self._create(config, name).to_dict()

    def destroy(self, deployment):
        """Destroy the deployment.

        :param deployment: UUID or name of the deployment
        """

        deploy = objects.Deployment.get(deployment)

        deploy.env_obj.destroy(skip_cleanup=True)
        deploy.env_obj.delete()

    def recreate(self, deployment, config=None):
        """Performs a cleanup and then makes a deployment again.

        :param deployment: UUID or name of the deployment
        :param config: an optional dict with deployment config to update before
                       redeploy
        """
        raise exceptions.RallyException("Sorry, but recreate method of "
                                        "deployments is temporary disabled.")

    def _get(self, deployment):
        """Get the deployment.

        :param deployment: UUID or name of the deployment
        :returns: Deployment instance
        """
        return objects.Deployment.get(deployment)

    def get(self, deployment):
        return self._get(deployment).to_dict()

    def list(self, status=None, parent_uuid=None, name=None):
        """Get the deployments list.

        :returns: Deployment list
        """
        return [deployment.to_dict() for deployment in
                objects.Deployment.list(status, parent_uuid, name)]

    def check(self, deployment):
        """Check keystone authentication and list all available services.

        :param deployment: UUID of deployment
        :returns: Service list
        """
        env = self._get(deployment).env_obj

        result = {}

        for p, res in env.check_health().items():
            name = "openstack" if p == "existing@openstack" else p
            if not res["available"]:
                # NOTE(andreykurilin): the old behavior supports 2 keys
                #   for storing errors: admin_error and user_error.
                #   Since admin/users is not mandatory thing for new design
                #   of Platforms, let's put platform error to "admin_error"
                key = "admin_error"
                if name == "openstack":
                    if res["message"].startswith("Bad user creds"):
                        key = "user_error"

                if "traceback" in res:
                    # NOTE(andreykurilin): the last not null line in traceback
                    #   includes Exception cls with a message. By parsing it,
                    #   we can get etype.
                    trace = res["traceback"].split("\n")
                    last_line = [l for l in trace if l][-1]
                    etype, _msg = last_line.split(":", 1)
                else:
                    etype = "n/a"
                result[name] = [
                    {
                        key: {
                            "etype": etype,
                            "msg": res["message"],
                            "trace": res.get("traceback", "n/a")
                        },
                        "services": []
                    }
                ]
            else:
                if name == "openstack":
                    services = env.get_info()[p]["info"]["services"]
                    # backward compatibility
                    for s in services:
                        if "name" not in s:
                            s["name"] = "__unknown__"

                    result[name] = [{"services": services}]
                else:
                    # NOTE(andreykurilin): the info method of platforms allows
                    #   to return whatever single platform wants, i.e
                    #   Platform X can return just a version and no services
                    #   at all. Checking for 'services' key there is not a
                    #   solution, since the value of it can have the format
                    #   which differs from openstack-like (the old design of
                    #   Deployment component)
                    result[name] = [{"services": []}]

        return result


class _Task(APIGroup):

    TASK_SCHEMA = objects.task.TASK_SCHEMA

    def list(self, **filters):
        return [task.to_dict() for task in objects.Task.list(**filters)]

    def get(self, task_id, detailed=False):
        """Get task data

        :param task_id: Task UUID
        :param detailed: whether return detailed information(including
            subtasks and workloads) or not.
        """
        return objects.Task.get(task_id, detailed=detailed).to_dict()

    # TODO(andreykurilin): move it to some kind of utils
    def render_template(self, task_template, template_dir="./", **kwargs):
        """Render jinja2 task template to Rally input task.

        :param task_template: String that contains template
        :param template_dir: The path of directory contain template files
        :param kwargs: Dict with template arguments
        :returns: rendered template str
        """

        def is_really_missing(mis, task_template):
            # NOTE(boris-42): Removing variables that have default values from
            #                 missing. Construction that won't be properly
            #                 checked is {% set x = x or 1}
            if re.search(mis.join([r"{%\s*set\s+", r"\s*=\s*", r"[^\w]+"]),
                         task_template):
                return False
            # NOTE(jlk): Also check for a default filter which can show up as
            #            a missing variable
            if re.search(mis + r"\s*\|\s*default\(", task_template):
                return False
            return True

        # NOTE(boris-42): We have to import builtins to get the full list of
        #                 builtin functions (e.g. range()). Unfortunately,
        #                 __builtins__ doesn't return them (when it is not
        #                 main module)
        import builtins

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir, encoding="utf8"))
        env.globals.update(self.create_template_functions())
        ast = env.parse(task_template)
        # NOTE(Julia Varigina):
        # Bug in jinja2.meta.find_undeclared_variables
        #
        # The method shows inconsistent behavior:
        # it does not return undeclared variables that appear
        # in included templates only (via {%- include "some_template.yaml"-%})
        # and in the same time is declared in jinja2.Environment.globals.
        #
        # This is different for undeclared variables that appear directly
        # in task_template. The method jinja2.meta.find_undeclared_variables
        # returns an undeclared variable that is used in task_template
        # and is set in jinja2.Environment.globals.
        #
        # Despite this bug, jinja resolves values
        # declared in jinja2.Environment.globals for both types of undeclared
        # variables and successfully renders templates in both cases.
        required_kwargs = jinja2.meta.find_undeclared_variables(ast)
        missing = (set(required_kwargs) - set(kwargs) - set(dir(builtins))
                   - set(env.globals))
        real_missing = [mis for mis in missing
                        if is_really_missing(mis, task_template)]
        if real_missing:
            multi_msg = "Please specify next template task arguments: %s"
            single_msg = "Please specify template task argument: %s"

            raise TypeError((len(real_missing) > 1 and multi_msg or single_msg)
                            % ", ".join(real_missing))

        render_template = env.from_string(task_template).render(**kwargs)
        return render_template

    def create_template_functions(self):

        def template_min(int1, int2):
            return min(int1, int2)

        def template_max(int1, int2):
            return max(int1, int2)

        def template_round(float1):
            return int(round(float1))

        def template_ceil(float1):
            import math
            return int(math.ceil(float1))

        return {"min": template_min, "max": template_max,
                "ceil": template_ceil, "round": template_round}

    def create(self, deployment, tags=None):
        """Create a task without starting it.

        Task is a list of subtasks that are called one by one, results of
        execution are stored into DB.

        Every subtask is sort of test case which is created by combination
        of scenario, runner, contexts, sla and hoooks plugins.

        :param deployment: UUID or name of the deployment
        :param tags: a list of tags for this task
        :returns: Task object
        """
        deployment = objects.Deployment.get(deployment)
        if deployment["status"] != consts.DeployStatus.DEPLOY_FINISHED:
            raise exceptions.DeploymentNotFinishedStatus(
                name=deployment["name"],
                uuid=deployment["uuid"],
                status=deployment["status"])

        return objects.Task(env_uuid=deployment["uuid"],
                            tags=tags).to_dict()

    def validate(self, deployment, config, task_instance=None, task=None):
        """Validate a task config against specified deployment.

        :param deployment: UUID or name of the deployment (will be ignored in
            case of transmitting task_instance or task arguments)
        :param config: a dict with a task configuration
        :param task_instance: DEPRECATED. Use "task" argument to transmit task
            uuid instead
        """
        if task_instance is not None:
            LOG.warning("Transmitting task object in `task validate` is "
                        "deprecated since Rally 0.10. To use pre-created "
                        "task, transmit task UUID instead via `task` "
                        "argument.")
            task = objects.Task.get(task_instance["uuid"])
            deployment = task["deployment_uuid"]
        elif task:
            task = objects.Task.get(task)
            deployment = task["deployment_uuid"]
        else:
            task = objects.Task(env_uuid=deployment, temporary=True)
        deployment = objects.Deployment.get(deployment)

        try:
            config = task_cfg.TaskConfig(config)
        except exceptions.InvalidTaskException:
            # it is a proper formed exception, nothing to do
            raise
        except Exception as e:
            if logging.is_debug():
                LOG.exception("Unexpected error had happened")
            raise exceptions.InvalidTaskException(str(e))

        engine.TaskEngine(config, task, deployment.env_obj).validate()

    def start(self, deployment, config, task=None, abort_on_sla_failure=False):
        """Validate and start a task.

        Task is a list of subtasks that are called one by one, results of
        execution are stored in DB.

        :param deployment: UUID or name of the deployment (will be ignored in
            case of transmitting existing task)
        :param config: a dict with a task configuration
        :param task: Task UUID to use pre-created task. If None, new task will
            be created
        :param abort_on_sla_failure: If set to True, the task execution is
                                     stop when any of SLA checks fails
        """
        if task and isinstance(task, objects.Task):
            LOG.warning("Transmitting task object in `task start` is "
                        "deprecated since Rally 0.10. To use pre-created "
                        "task, transmit task UUID instead.")
            if task.is_temporary:
                raise ValueError(
                    "Unable to run a temporary task. Please check your code.")
            task = objects.Task.get(task["uuid"])
        elif task is not None:
            task = objects.Task.get(task)

        if task is not None:
            deployment = task["deployment_uuid"]

        deployment = objects.Deployment.get(deployment)
        if deployment["status"] != consts.DeployStatus.DEPLOY_FINISHED:
            raise exceptions.DeploymentNotFinishedStatus(
                name=deployment["name"],
                uuid=deployment["uuid"],
                status=deployment["status"])

        try:
            config = task_cfg.TaskConfig(config)
        except exceptions.InvalidTaskException:
            # it is a proper formed exception, nothing to do
            raise
        except Exception as e:
            if logging.is_debug():
                LOG.exception("Unexpected error had happened")
            raise exceptions.InvalidTaskException(str(e))

        if task is None:
            task = objects.Task(deployment_uuid=deployment["uuid"],
                                title=config.title,
                                description=config.description)

        task_engine = engine.TaskEngine(
            config, task, deployment.env_obj,
            abort_on_sla_failure=abort_on_sla_failure)

        task_engine.validate()

        LOG.info("Task %s input file is valid." % task["uuid"])
        LOG.info("Run Task %s against Deployment %s"
                 % (task["uuid"], deployment["uuid"]))

        task_engine.run()

        return task["uuid"], task.get_status(task["uuid"])

    def abort(self, task_uuid, soft=False, wait=False, **kwargs):
        """Abort running task.

        :param task_uuid: The UUID of the task
        :type task_uuid: str
        :param soft: If set to True, task should be aborted after execution of
                     current scenario, otherwise as soon as possible before
                     all the scenario iterations finish [Default: False]
        :type soft: bool
        :param wait: wait until task stops [Default: False]
        :type wait: bool
        """
        if kwargs:
            if len(kwargs) != 1 or "async" not in kwargs:
                raise TypeError("API method task.abort accept only one "
                                "argument 'async' (which is deprecated in "
                                "favor of 'wait').")
            elif "async" in kwargs:
                LOG.warning("The argument 'async' of API method task.abort is "
                            "deprecated since Rally 1.1.0 in favor of new "
                            "argument 'wait' which doesn't conflict with a "
                            "reserved keywords in python 3.7")
                wait = not kwargs["async"]

        if wait:
            current_status = objects.Task.get_status(task_uuid)
            if current_status in objects.Task.NOT_IMPLEMENTED_STAGES_FOR_ABORT:
                LOG.info(
                    "Task status is '%s' waiting until it became 'running'"
                    % current_status)
                while (current_status in
                       objects.Task.NOT_IMPLEMENTED_STAGES_FOR_ABORT):
                    time.sleep(1)
                    current_status = objects.Task.get_status(task_uuid)

        objects.Task.get(task_uuid).abort(soft=soft)

        if wait:
            LOG.info("Waiting until the task stops.")
            finished_stages = [consts.TaskStatus.ABORTED,
                               consts.TaskStatus.FINISHED,
                               consts.TaskStatus.CRASHED]
            while objects.Task.get_status(task_uuid) not in finished_stages:
                time.sleep(1)

    def delete(self, task_uuid, force=False):
        """Deletes all task data from database.

        :param task_uuid: The UUID of the task
        :param force: If set to True, then delete the task despite to the
                      status
        :raises DBConflict: when the status of the task is not
                            in FINISHED, FAILED or ABORTED and
                            the force argument is not True
        :raises DBRecordNotFound: when task doesn't exist
        """
        if force:
            objects.Task.delete_by_uuid(task_uuid, status=None)
        elif objects.Task.get_status(task_uuid) in (
                consts.TaskStatus.ABORTED,
                consts.TaskStatus.FINISHED,
                consts.TaskStatus.CRASHED):
            objects.Task.delete_by_uuid(task_uuid, status=None)
        else:
            objects.Task.delete_by_uuid(
                task_uuid, status=consts.TaskStatus.FINISHED)

    def import_results(self, deployment, task_results, tags=None):
        """Import json results of a task into rally database"""
        deployment = objects.Deployment.get(deployment)
        if deployment["status"] != consts.DeployStatus.DEPLOY_FINISHED:
            raise exceptions.DeploymentNotFinishedStatus(
                name=deployment["name"],
                uuid=deployment["uuid"],
                status=deployment["status"])

        task_inst = objects.Task(env_uuid=deployment["uuid"],
                                 tags=tags)
        task_inst.update_status(consts.TaskStatus.RUNNING)
        for subtask in task_results["subtasks"]:
            subtask_obj = task_inst.add_subtask(title=subtask.get("title"))
            for workload in subtask["workloads"]:
                for data in workload["data"]:
                    if not task_inst.result_has_valid_schema(data):
                        raise exceptions.RallyException(
                            "Task %s is trying to import "
                            "results in wrong format" % task_inst["uuid"])

                workload_obj = subtask_obj.add_workload(
                    name=workload["name"], description=workload["description"],
                    position=workload["position"], runner=workload["runner"],
                    runner_type=workload["runner_type"],
                    contexts=workload["contexts"], hooks=workload["hooks"],
                    sla=workload["sla"], args=workload["args"])

                chunk_size = CONF.raw_result_chunk_size
                workload_data_count = 0
                while len(workload["data"]) > chunk_size:
                    results_chunk = workload["data"][:chunk_size]
                    workload["data"] = workload["data"][chunk_size:]
                    results_chunk.sort(key=lambda x: x["timestamp"])
                    workload_obj.add_workload_data(workload_data_count,
                                                   {"raw": results_chunk})
                    workload_data_count += 1

                workload_obj.add_workload_data(workload_data_count,
                                               {"raw": workload["data"]})
                workload_obj.set_results(
                    sla_results=workload["sla_results"].get("sla"),
                    hooks_results=workload["hooks"],
                    start_time=workload["start_time"],
                    full_duration=workload["full_duration"],
                    load_duration=workload["load_duration"],
                    contexts_results=workload["contexts_results"])
            subtask_obj.update_status(consts.SubtaskStatus.FINISHED)
        task_inst.update_status(consts.SubtaskStatus.FINISHED)

        LOG.info("Task results have been successfully imported.")

        return task_inst.to_dict()

    def export(self, tasks, output_type, output_dest=None):
        """Generate a report for a task or a few tasks.

        :param tasks: List of tasks UUIDs or tasks results
        :param output_type: Plugin name of task exporter
        :param output_dest: Destination for task report
        """

        tasks_results = []
        tasks = tasks or []
        for task in tasks:
            if isinstance(task, dict):
                tasks_results.append(task)
            else:
                tasks_results.append(self.get(task_id=task, detailed=True))

        errors = texporter.TaskExporter.validate(
            output_type, context={}, config={},
            # wrap destination to a dict to allow extending options in future
            plugin_cfg={"destination": output_dest},
            vtype="syntax"
        )
        if errors:
            raise exceptions.ValidationError("\n".join(errors))

        reporter_cls = texporter.TaskExporter.get(output_type)

        LOG.info("Building '%s' report for the following task(s): '%s'."
                 % (output_type,
                    "', '".join([task["uuid"] for task in tasks_results])))
        result = texporter.TaskExporter.make(reporter_cls,
                                             tasks_results,
                                             output_dest,
                                             api=self.api)
        LOG.info("The report has been successfully built.")
        return result


class _Verifier(APIGroup):

    def list_plugins(self, platform=None):
        """List all plugins for verifiers management.

        :param platform: Verifier plugin platform
        """
        return [{"name": p.get_name(),
                 "platform": p.get_platform(),
                 "description": p.get_info()["title"],
                 "location": "%s.%s" % (p.__module__, p.__name__)}
                for p in vmanager.VerifierManager.get_all(platform=platform)]

    def create(self, name, vtype, platform=None, source=None, version=None,
               system_wide=False, extra_settings=None):
        """Create a verifier.

        :param name: Verifier name
        :param vtype: Verifier plugin name
        :param platform: Verifier plugin platform. Should be specified when
                          there are two verifier plugins with equal names but
                          in different platforms
        :param source: Path or URL to the repo to clone verifier from
        :param version: Branch, tag or commit ID to checkout before
                        verifier installation
        :param system_wide: Whether or not to use the system-wide environment
                            for verifier instead of a virtual environment
        :param extra_settings: Extra installation settings for verifier
        """
        # check that the specified verifier type exists
        vmanager.VerifierManager.get(vtype, platform=platform)

        LOG.info("Creating verifier '%s'." % name)

        try:
            verifier = self._get(name)
        except exceptions.DBRecordNotFound:
            verifier = objects.Verifier.create(
                name=name, source=source, system_wide=system_wide,
                version=version, vtype=vtype, platform=platform,
                extra_settings=extra_settings)
        else:
            raise exceptions.RallyException(
                "Verifier with name '%s' already exists! Please, specify "
                "another name for verifier and try again." % verifier.name)

        properties = {}
        properties["platform"] = platform or verifier.manager.get_platform()

        default_source = verifier.manager._meta_get("default_repo")
        if not source and default_source:
            properties["source"] = default_source

        if properties:
            verifier.update_properties(**properties)

        verifier.update_status(consts.VerifierStatus.INSTALLING)
        try:
            verifier.manager.install()
        except Exception:
            verifier.update_status(consts.VerifierStatus.FAILED)
            raise
        verifier.update_status(consts.VerifierStatus.INSTALLED)

        LOG.info("Verifier %s has been successfully created!" % verifier)

        return verifier.uuid

    def _get(self, verifier_id):
        """Get a verifier.

        :param verifier_id: Verifier name or UUID
        """
        return objects.Verifier.get(verifier_id)

    def get(self, verifier_id):
        return self._get(verifier_id).to_dict()

    def _list(self, status=None):
        """List all verifiers.

        :param status: Status to filter verifiers by
        """
        return objects.Verifier.list(status)

    def list(self, status=None):
        return [item.to_dict() for item in self._list(status)]

    def delete(self, verifier_id, deployment_id=None, force=False):
        """Delete a verifier.

        :param verifier_id: Verifier name or UUID
        :param deployment_id: Deployment name or UUID. If specified,
                              only the deployment-specific data will be deleted
                              for verifier
        :param force: Delete all stored verifier verifications.
                      If deployment_id specified, only verifications of this
                      deployment will be deleted
        """
        verifier = self._get(verifier_id)
        verifications = self.api.verification.list(
            verifier_id=verifier_id,
            deployment_id=deployment_id)
        if verifications:
            d_msg = ((" for deployment '%s'" % deployment_id)
                     if deployment_id else "")
            if force:
                LOG.info("Deleting all verifications created by verifier %s%s."
                         % (verifier, d_msg))
                for verification in verifications:
                    self.api.verification.delete(
                        verification_uuid=verification["uuid"])
            else:
                raise exceptions.RallyException(
                    "Failed to delete verifier {0} because there are stored "
                    "verifier verifications{1}! Please, make sure that they "
                    "are not important to you. Use 'force' flag if you would "
                    "like to delete verifications{1} as well."
                    .format(verifier, d_msg))

        if deployment_id:
            LOG.info("Deleting deployment-specific data for verifier %s."
                     % verifier)
            verifier.set_env(deployment_id)
            verifier.manager.uninstall()
            LOG.info("Deployment-specific data has been successfully deleted!")
        else:
            LOG.info("Deleting verifier %s." % verifier)
            verifier.manager.uninstall(full=True)
            objects.Verifier.delete(verifier_id)
            LOG.info("Verifier has been successfully deleted!")

    def update(self, verifier_id, system_wide=None, version=None,
               update_venv=False):
        """Update a verifier.

        :param verifier_id: Verifier name or UUID
        :param system_wide: Switch to using the system-wide environment
        :param version: Branch, tag or commit ID to checkout
        :param update_venv: Update the virtual environment for verifier
        """
        if system_wide is None and version is None and not update_venv:
            # nothing to update
            raise exceptions.RallyException(
                "At least one of the following parameters should be "
                "specified: 'system_wide', 'version', 'update_venv'.")

        verifier = self._get(verifier_id)
        LOG.info("Updating verifier %s." % verifier)

        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to update verifier %s because verifier is in '%s' "
                "status, but should be in '%s'." % (
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        system_wide_in_use = (
            system_wide or (system_wide is None and verifier.system_wide))
        if update_venv and system_wide_in_use:
            raise exceptions.RallyException(
                "It is impossible to update the virtual environment for "
                "verifier %s when it uses the system-wide environment."
                % verifier)

        # store original status to set it again after updating or rollback
        original_status = verifier.status
        verifier.update_status(consts.VerifierStatus.UPDATING)

        properties = {}  # store new verifier properties to update old ones

        sw_is_checked = False

        if version:
            properties["version"] = version

            backup = utils.BackupHelper()
            rollback_msg = ("Failed to update verifier %s. It has been "
                            "rollbacked to the previous state." % verifier)
            backup.add_rollback_action(LOG.info, rollback_msg)
            backup.add_rollback_action(verifier.update_status, original_status)
            with backup(verifier.manager.repo_dir):
                verifier.manager.checkout(version)

            if system_wide_in_use:
                verifier.manager.check_system_wide()
                sw_is_checked = True

        if system_wide is not None:
            if system_wide == verifier.system_wide:
                LOG.info(
                    "Verifier %s is already switched to system_wide=%s. "
                    "Nothing will be changed."
                    % (verifier, verifier.system_wide))
            else:
                properties["system_wide"] = system_wide
                if not system_wide:
                    update_venv = True  # we need to install a virtual env
                else:
                    # NOTE(andreykurilin): should we remove previously created
                    #   virtual environment?!
                    if not sw_is_checked:
                        verifier.manager.check_system_wide()

        if update_venv:
            backup = utils.BackupHelper()
            rollback_msg = ("Failed to update the virtual environment for "
                            "verifier %s. It has been rollbacked to the "
                            "previous state." % verifier)
            backup.add_rollback_action(LOG.info, rollback_msg)
            backup.add_rollback_action(verifier.update_status, original_status)
            with backup(verifier.manager.venv_dir):
                verifier.manager.install_venv()

        properties["status"] = original_status  # change verifier status back
        verifier.update_properties(**properties)

        LOG.info("Verifier %s has been successfully updated!" % verifier)

        return verifier.uuid

    def configure(self, verifier, deployment_id, extra_options=None,
                  reconfigure=False):
        """Configure a verifier.

        :param verifier: Verifier object or (name or UUID)
        :param deployment_id: Deployment name or UUID
        :param extra_options: Extend verifier configuration with extra options
        :param reconfigure: Reconfigure verifier
        """
        if not isinstance(verifier, objects.Verifier):
            verifier = self._get(verifier)
        verifier.set_env(deployment_id)
        LOG.info("Configuring verifier %s for deployment '%s' (UUID=%s)."
                 % (verifier,
                    verifier.deployment["name"],
                    verifier.deployment["uuid"]))

        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to configure verifier %s for deployment '%s' "
                "(UUID=%s) because verifier is in '%s' status, but should be "
                "in '%s'." % (verifier, verifier.deployment["name"],
                              verifier.deployment["uuid"], verifier.status,
                              consts.VerifierStatus.INSTALLED))

        msg = ("Verifier %s has been successfully configured for deployment "
               "'%s' (UUID=%s)!" % (verifier, verifier.deployment["name"],
                                    verifier.deployment["uuid"]))
        vm = verifier.manager
        if vm.is_configured():
            LOG.info("Verifier is already configured!")
            if not reconfigure:
                if not extra_options:
                    return vm.get_configuration()
                else:
                    # Just add extra options to the config file.
                    if logging.is_debug():
                        LOG.debug("Adding the following extra options: %s "
                                  "to verifier configuration." % extra_options)
                    else:
                        LOG.info(
                            "Adding extra options to verifier configuration.")
                    vm.extend_configuration(extra_options)
                    LOG.info(msg)
                    return vm.get_configuration()

            LOG.info("Reconfiguring verifier.")

        raw_config = vm.configure(extra_options=extra_options)

        LOG.info(msg)

        return raw_config

    def override_configuration(self, verifier_id, deployment_id,
                               new_configuration):
        """Override verifier configuration (e.g., rewrite the config file).

        :param verifier_id: Verifier name or UUID
        :param deployment_id: Deployment name or UUID
        :param new_configuration: New configuration for verifier
        """
        verifier = self._get(verifier_id)
        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to override verifier configuration for deployment "
                "'%s' (UUID=%s) because verifier %s is in '%s' status, but "
                "should be in '%s'." % (
                    verifier.deployment["name"], verifier.deployment["uuid"],
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        verifier.set_env(deployment_id)
        LOG.info("Overriding configuration of verifier %s for deployment '%s' "
                 "(UUID=%s)."
                 % (verifier,
                    verifier.deployment["name"], verifier.deployment["uuid"]))
        verifier.manager.override_configuration(new_configuration)
        LOG.info("Configuration of verifier %s has been successfully "
                 "overridden for deployment '%s' (UUID=%s)!"
                 % (verifier,
                    verifier.deployment["name"], verifier.deployment["uuid"]))

    def list_tests(self, verifier_id, pattern=""):
        """List all verifier tests.

        :param verifier_id: Verifier name or UUID
        :param pattern: Pattern which will be used for matching
        """
        verifier = self._get(verifier_id)
        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to list verifier tests because verifier %s is in '%s' "
                "status, but should be in '%s'." % (
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        if pattern:
            verifier.manager.validate_args({"pattern": pattern})

        return verifier.manager.list_tests(pattern)

    def add_extension(self, verifier_id, source, version=None,
                      extra_settings=None):
        """Add a verifier extension.

        :param verifier_id: Verifier name or UUID
        :param source: Path or URL to the repo to clone verifier extension from
        :param version: Branch, tag or commit ID to checkout before
                        installation of the verifier extension
        :param extra_settings: Extra installation settings for verifier
                               extension
        """
        verifier = self._get(verifier_id)
        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to add verifier extension because verifier %s "
                "is in '%s' status, but should be in '%s'." % (
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        LOG.info("Adding extension for verifier %s." % verifier)

        # store original status to rollback it after failure
        original_status = verifier.status
        verifier.update_status(consts.VerifierStatus.EXTENDING)
        try:
            verifier.manager.install_extension(source, version=version,
                                               extra_settings=extra_settings)
        finally:
            verifier.update_status(original_status)

        LOG.info("Extension for verifier %s has been successfully added!"
                 % verifier)

    def list_extensions(self, verifier_id):
        """List all verifier extensions.

        :param verifier_id: Verifier name or UUID
        """
        verifier = self._get(verifier_id)
        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to list verifier extensions because verifier %s "
                "is in '%s' status, but should be in '%s.'" % (
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        return verifier.manager.list_extensions()

    def delete_extension(self, verifier_id, name):
        """Delete a verifier extension.

        :param verifier_id: Verifier name or UUID
        :param name: Verifier extension name
        """
        verifier = self._get(verifier_id)
        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to delete verifier extension because verifier %s "
                "is in '%s' status, but should be in '%s'." % (
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        LOG.info("Deleting extension for verifier %s." % verifier)
        verifier.manager.uninstall_extension(name)
        LOG.info("Extension for verifier %s has been successfully deleted!"
                 % verifier)


class _Verification(APIGroup):

    def start(self, verifier_id, deployment_id, tags=None, **run_args):
        """Start a verification.

        :param verifier_id: Verifier name or UUID
        :param deployment_id: Deployment name or UUID
        :param tags: List of tags to assign them to verification
        :param run_args: Dictionary with run arguments for verification
        """
        # TODO(ylobankov): Add an ability to skip tests by specifying only test
        #                  names (without test IDs). Also, it would be nice to
        #                  skip the whole test suites. For example, all tests
        #                  in the class or module.

        deployment = objects.Deployment.get(deployment_id)

        if deployment["status"] != consts.DeployStatus.DEPLOY_FINISHED:
            raise exceptions.DeploymentNotFinishedStatus(
                name=deployment["name"],
                uuid=deployment["uuid"],
                status=deployment["status"])

        verifier = self.api.verifier._get(verifier_id)
        if verifier.status != consts.VerifierStatus.INSTALLED:
            raise exceptions.RallyException(
                "Failed to start verification because verifier %s is in '%s' "
                "status, but should be in '%s'." % (
                    verifier, verifier.status, consts.VerifierStatus.INSTALLED)
            )

        verifier.set_env(deployment_id)
        if not verifier.manager.is_configured():
            self.api.verifier.configure(verifier=verifier,
                                        deployment_id=deployment_id)

        # TODO(andreykurilin): save validation results to db
        verifier.manager.validate(run_args)

        verification = objects.Verification.create(
            verifier_id=verifier_id, deployment_id=deployment_id, tags=tags,
            run_args=run_args)
        LOG.info("Starting verification (UUID=%s) for deployment '%s' "
                 "(UUID=%s) by verifier %s."
                 % (verification.uuid,
                    verifier.deployment["name"],
                    verifier.deployment["uuid"],
                    verifier))
        verification.update_status(consts.VerificationStatus.RUNNING)

        context = {"config": verifier.manager._meta_get("context"),
                   "run_args": run_args,
                   "verification": verification,
                   "verifier": verifier}
        try:
            with vcontext.ContextManager(context):
                results = verifier.manager.run(context)
        except Exception as e:
            verification.set_error(e)
            raise

        # TODO(ylobankov): Check that verification exists in the database
        #                  because users may delete verification before tests
        #                  finish.
        verification.finish(results.totals, results.tests)

        LOG.info("Verification (UUID=%s) has been successfully finished for "
                 "deployment '%s' (UUID=%s)!"
                 % (verification.uuid,
                    verifier.deployment["name"], verifier.deployment["uuid"]))

        return {"verification": verification.to_dict(),
                "totals": results.totals,
                "tests": results.tests}

    def rerun(self, verification_uuid, deployment_id=None, failed=False,
              tags=None, concurrency=0):
        """Rerun tests from a verification.

        :param verification_uuid: Verification UUID
        :param deployment_id: Deployment name or UUID
        :param failed: Rerun only failed tests
        :param tags: List of tags to assign them to verification
        :param concurrency: The number of processes to use to run verifier
            tests
        """
        # TODO(ylobankov): Improve this method in the future: put some
        #                  information about re-run in run_args.
        run_args = {}
        if concurrency:
            run_args["concurrency"] = concurrency

        verification = self._get(verification_uuid)
        tests = verification.tests

        if failed:
            tests = [t for t, r in tests.items() if r["status"] == "fail"]
            if not tests:
                raise exceptions.RallyException(
                    "There are no failed tests from verification (UUID=%s)."
                    % verification_uuid)
        else:
            # py2 and py3
            tests = list(tests.keys())

        deployment = (deployment_id if deployment_id
                      else verification.deployment_uuid)
        deployment = self.api.deployment.get(deployment=deployment)
        LOG.info("Re-running %stests from verification (UUID=%s) for "
                 "deployment '%s' (UUID=%s)."
                 % ("failed " if failed else "",
                    verification.uuid,
                    deployment["name"], deployment["uuid"]))
        return self.start(verifier_id=verification.verifier_uuid,
                          deployment_id=deployment["uuid"],
                          load_list=tests, tags=tags, **run_args)

    def _get(self, verification_uuid):
        """Get a verification.

        :param verification_uuid: Verification UUID
        """
        return objects.Verification.get(verification_uuid)

    def get(self, verification_uuid):
        return self._get(verification_uuid).to_dict()

    def list(self, verifier_id=None, deployment_id=None,
             tags=None, status=None):
        """List all verifications.

        :param verifier_id: Verifier name or UUID
        :param deployment_id: Deployment name or UUID
        :param tags: Tags to filter verifications by
        :param status: Status to filter verifications by
        """
        return [item.to_dict() for item in objects.Verification.list(
            verifier_id, deployment_id=deployment_id,
            tags=tags, status=status)]

    def delete(self, verification_uuid):
        """Delete a verification.

        :param verification_uuid: Verification UUID
        """
        verification = self._get(verification_uuid)
        LOG.info("Deleting verification (UUID=%s)." % verification.uuid)
        verification.delete()
        LOG.info("Verification has been successfully deleted!")

    def report(self, uuids, output_type, output_dest=None):
        """Generate a report for a verification or a few verifications.

        :param uuids: List of verifications UUIDs
        :param output_type: Plugin name of verification reporter
        :param output_dest: Destination for verification report
        """
        verifications = [self._get(uuid) for uuid in uuids]

        reporter_cls = vreporter.VerificationReporter.get(output_type)
        reporter_cls.validate(output_dest)

        LOG.info("Building '%s' report for the following verification(s): '%s'"
                 % (output_type, "', '".join(uuids)))
        result = vreporter.VerificationReporter.make(reporter_cls,
                                                     verifications,
                                                     output_dest)
        LOG.info("The report has been successfully built.")
        return result

    def import_results(self, verifier_id, deployment_id, data, **run_args):
        """Import results of a test run into Rally database.

        :param verifier_id: Verifier name or UUID
        :param deployment_id: Deployment name or UUID
        :param data: Results data of a test run to import
        :param run_args: Dictionary with run arguments
        """
        # TODO(aplanas): Create an external deployment if this is missing, as
        # required in the blueprint [1].
        # [1] https://blueprints.launchpad.net/rally/+spec/verification-import

        verifier = self.api.verifier._get(verifier_id)
        verifier.set_env(deployment_id)
        LOG.info("Importing test results into a new verification for "
                 "deployment '%s' (UUID=%s), using verifier %s."
                 % (verifier.deployment["name"],
                    verifier.deployment["uuid"],
                    verifier))

        verifier.manager.validate_args(run_args)

        verification = objects.Verification.create(verifier_id,
                                                   deployment_id=deployment_id,
                                                   run_args=run_args)
        verification.update_status(consts.VerificationStatus.RUNNING)

        try:
            results = verifier.manager.parse_results(data)
        except Exception as e:
            verification.set_failed(e)
            raise
        verification.finish(results.totals, results.tests)

        LOG.info("Test results have been successfully imported.")

        return {"verification": verification.to_dict(),
                "totals": results.totals,
                "tests": results.tests}


class API(object):

    CONFIG_SEARCH_PATHS = [sys.prefix + "/etc/rally", "~/.rally", "/etc/rally"]
    CONFIG_FILE_NAME = "rally.conf"

    def __init__(self, config_file=None, config_args=None,
                 plugin_paths=None, skip_db_check=False):
        """Initialize Rally API instance

        :param config_file: Path to rally configuration file. If None, default
                            path will be selected
        :type config_file: str
        :param config_args: Arguments for initialization current configuration
        :type config_args: list
        :param plugin_paths: Additional custom plugin locations
        :type plugin_paths: list
        :param skip_db_check: Allows to skip db revision check
        :type skip_db_check: bool
        """

        try:
            config_files = ([config_file] if config_file else
                            self._default_config_file())
            CONF(config_args or [],
                 project="rally",
                 version=rally_version.version_string(),
                 default_config_files=config_files)
            CONF.set_default("use_stderr", True)

            logging.setup("rally")
            if not CONF.get("log_config_append"):
                # The below two lines are to disable noise from request module.
                # The standard way should be we make such lots of settings on
                # the root rally. However current oslo codes doesn't support
                # such interface. So I choose to use a 'hacking' way to avoid
                # INFO logs from request module where user didn't give specific
                # log configuration. And we could remove this hacking after
                # oslo.log has such interface.
                LOG.debug(
                    "INFO logs from urllib3 and requests module are hide.")
                requests_log = logging.getLogger("requests").logger
                requests_log.setLevel(logging.WARNING)
                urllib3_log = logging.getLogger("urllib3").logger
                urllib3_log.setLevel(logging.WARNING)

                LOG.debug("urllib3 insecure warnings are hidden.")
                for warning in ("InsecurePlatformWarning",
                                "SNIMissingWarning",
                                "InsecureRequestWarning"):
                    warning_cls = getattr(urllib3.exceptions, warning, None)
                    if warning_cls is not None:
                        urllib3.disable_warnings(warning_cls)

            # NOTE(wtakase): This is for suppressing boto error logging.
            LOG.debug("ERROR log from boto module is hide.")
            boto_log = logging.getLogger("boto").logger
            boto_log.setLevel(logging.CRITICAL)

            # Set alembic log level to ERROR
            alembic_log = logging.getLogger("alembic").logger
            alembic_log.setLevel(logging.ERROR)

        except cfg.ConfigFilesNotFoundError as e:
            cfg_files = e.config_files
            raise exceptions.RallyException(
                "Failed to read configuration file(s): %s" % cfg_files)

        # Check that db is upgraded to the latest revision
        if not skip_db_check:
            self.check_db_revision()

        # Load plugins
        plugin_paths = plugin_paths or []
        if "plugin_paths" in CONF:
            plugin_paths.extend(CONF.get("plugin_paths") or [])
        for path in plugin_paths:
            discover.load_plugins(path)

        # NOTE(andreykurilin): There is no reason to auto-discover API's. We
        # have only 4 classes, so let's do it in good old way - hardcode them:)
        self._deployment = _Deployment(self)
        self._task = _Task(self)
        self._verifier = _Verifier(self)
        self._verification = _Verification(self)

    def _default_config_file(self):
        for path in self.CONFIG_SEARCH_PATHS:
            abspath = os.path.abspath(os.path.expanduser(path))
            fpath = os.path.join(abspath, self.CONFIG_FILE_NAME)
            if os.path.isfile(fpath):
                return [fpath]

    def check_db_revision(self):
        rev = rally_version.database_revision()

        # Check that db exists
        if rev["revision"] is None:
            raise exceptions.RallyException(
                "Database is missing. Create database by command "
                "`rally db create'")

        # Check that db is updated
        if rev["revision"] != rev["current_head"]:
            raise exceptions.RallyException((
                "Database seems to be outdated. Run upgrade from "
                "revision %(revision)s to %(current_head)s by command "
                "`rally db upgrade'") % rev)

    def _request(self, path, method, **kwargs):
        headers = {
            "RALLY-CLIENT-VERSION": rally_version.version_string(),
            "RALLY-API": "1.0"
        }
        response = requests.request(method, path,
                                    json=kwargs, headers=headers)
        if response.status_code != 200:
            raise exceptions.find_exception(response)

        return response.json(
            object_pairs_hook=collections.OrderedDict)["result"]

    @property
    def deployment(self):
        return self._deployment

    @property
    def task(self):
        return self._task

    @property
    def verifier(self):
        return self._verifier

    @property
    def verification(self):
        return self._verification

    @property
    def version(self):
        return 1



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from docutils.parsers import rst
import json
import re

from . import utils
from rally.common.plugin import discover
from rally.common.plugin import plugin
from rally.common import validation
from rally import plugins


JSON_SCHEMA_TYPES_MAP = {"boolean": "bool",
                         "string": "str",
                         "number": "float",
                         "integer": "int",
                         "array": "list",
                         "object": "dict"}


def process_jsonschema(schema):
    """Process jsonschema and make it looks like regular docstring."""

    if not schema:
        # nothing to parse
        return

    if "type" in schema:

        # str
        if schema["type"] == "string":
            doc = schema.get("description", "")
            if "pattern" in schema:
                doc += ("\n\nShould follow next pattern: %s." %
                        schema["pattern"])
            return {"doc": doc, "type": "str"}

        # int or float
        elif schema["type"] in ("integer", "number"):
            doc = schema.get("description", "")
            if "minimum" in schema:
                doc += "\n\nMin value: %s." % schema["minimum"]
            if "maximum" in schema:
                doc += "\n\nMax value: %s." % schema["maximum"]
            return {"doc": doc, "type": JSON_SCHEMA_TYPES_MAP[schema["type"]]}

        # bool or null
        elif schema["type"] in ("boolean", "null"):
            return {"doc": schema.get("description", ""),
                    "type": "bool" if schema["type"] == "boolean" else "null"}

        # list
        elif schema["type"] == "array":
            info = {"doc": schema.get("description", ""),
                    "type": "list"}

            if "items" in schema:
                if info["doc"]:
                    info["doc"] += "\n\n"
                info["doc"] += ("Elements of the list should follow format(s) "
                                "described below:\n\n")

                items = schema["items"]
                itype = None
                if "type" in items:
                    itype = JSON_SCHEMA_TYPES_MAP.get(items["type"],
                                                      items["type"])
                    info["doc"] += "- Type: %s. " % itype
                if "description" in items:
                    # add indention
                    desc = items["description"].split("\n")
                    info["doc"] += "\n  ".join(desc)

                new_schema = copy.copy(items)
                new_schema.pop("description", None)
                new_schema = json.dumps(new_schema, indent=4)
                new_schema = "\n     ".join(
                    new_schema.split("\n"))

                info["doc"] += ("\n  Format:\n\n"
                                "    .. code-block:: json\n\n"
                                "      %s\n" % new_schema)
            return info

        elif isinstance(schema["type"], list):
            # it can be too complicated for parsing... do not do it deeply
            return {"doc": schema.get("description", ""),
                    "type": "/".join(schema["type"])}

        # dict
        elif schema["type"] == "object":
            info = {"doc": schema.get("description", ""),
                    "type": "dict",
                    "parameters": []}
            required_parameters = schema.get("required", [])
            if "properties" in schema:
                for name in schema["properties"]:
                    if isinstance(schema["properties"][name], str):
                        pinfo = {"name": name,
                                 "type": schema["properties"][name],
                                 "doc": ""}
                    else:
                        pinfo = process_jsonschema(schema["properties"][name])
                        if name in required_parameters:
                            pinfo["required"] = True
                        pinfo["name"] = name
                    info["parameters"].append(pinfo)
            elif "patternProperties" in schema:
                info.pop("parameters", None)
                info["patternProperties"] = []
                for k, v in schema["patternProperties"].items():
                    info["patternProperties"].append(process_jsonschema(v))
                    info["patternProperties"][-1]["name"] = k
                    info["patternProperties"][-1]["type"] = "str"
            elif (not (set(schema.keys()) - {"type", "description", "$schema",
                                             "additionalProperties"})):
                # it is ok, schema accepts any object. nothing to add more
                pass
            elif "oneOf" in schema or "anyOf" in schema:
                # Example:
                #   SCHEMA = {"type": "object", "$schema": consts.JSON_SCHEMA,
                #       "oneOf": [{"properties": {"foo": {"type": "string"}}
                #                  "required": ["foo"],
                #                  "additionalProperties": False},
                #                 {"properties": {"bar": {"type": "string"}}
                #                  "required": ["bar"],
                #                  "additionalProperties": False},
                #
                schema_key = "oneOf" if "oneOf" in schema else "anyOf"
                schema_value = copy.deepcopy(schema.get(schema_key))

                for item in schema_value:
                    for k, v in schema.items():
                        if k not in (schema_key, "description"):
                            item[k] = v

                return {
                    "doc": schema.get("description", ""),
                    "type": "dict",
                    schema_key: [
                        process_jsonschema(item) for item in schema_value]
                }
            else:
                raise Exception("Failed to parse jsonschema: %s" % schema)

            if "definitions" in schema:
                info["definitions"] = schema["definitions"]
            return info
        else:
            raise Exception("Failed to parse jsonschema: %s" % schema)

    # enum
    elif "enum" in schema:
        doc = schema.get("description", "")
        doc += "\nSet of expected values: '%s'." % ("', '".join(
            [e or "None" for e in schema["enum"]]))
        return {"doc": doc}

    elif "anyOf" in schema:
        return {"doc": schema.get("description", ""),
                "anyOf": [process_jsonschema(i) for i in schema["anyOf"]]}

    elif "oneOf" in schema:
        return {"doc": schema.get("description", ""),
                "oneOf": [process_jsonschema(i) for i in schema["oneOf"]]}

    elif "$ref" in schema:
        return {"doc": schema.get("description", "n/a"),
                "ref": schema["$ref"]}
    else:
        raise Exception("Failed to parse jsonschema: %s" % schema)


CATEGORIES = {
    "Common": ["OS Client"],
    "Deployment": ["Engine", "Provider Factory"],
    "Task Component": ["Chart", "Context", "Hook Action", "Hook Trigger",
                       "Resource Type", "Task Exporter", "SLA", "Scenario",
                       "Scenario Runner", "Trigger", "Validator"],
    "Verification Component": ["Verifier Context", "Verification Reporter",
                               "Verifier Manager"]
}

# NOTE(andreykurilin): several bases do not have docstings at all, so it is
# redundant to display them
IGNORED_BASES = ["Resource Type", "OS Client", "Exporters"]


class PluginsReferenceDirective(rst.Directive):
    optional_arguments = 1
    option_spec = {"base_cls": str}

    def _make_arg_items(self, items, ref_prefix, description=None,
                        title="Parameters"):
        terms = []
        for item in items:
            iname = item.get("name", "") or item.pop("type")
            if "type" in item:
                iname += " (%s)" % item["type"]
            terms.append((iname, [item["doc"]]))
        return utils.make_definitions(title=title,
                                      ref_prefix=ref_prefix,
                                      terms=terms,
                                      descriptions=description)

    def _make_plugin_section(self, plugin_cls, base_name=None):
        section_name = plugin_cls.get_name()
        if base_name:
            section_name += " [%s]" % base_name
        section_obj = utils.section(section_name)

        info = plugin_cls.get_info()
        if info["title"]:
            section_obj.append(utils.paragraph(info["title"]))

        if info["description"]:
            section_obj.extend(utils.parse_text(info["description"]))

        if info["platform"]:
            section_obj.append(utils.paragraph(
                "**Platform**: %s" % info["platform"]))

        if base_name:
            ref_prefix = "%s-%s-" % (base_name, plugin_cls.get_name())
        else:
            ref_prefix = "%s-" % plugin_cls.get_name()

        if info["parameters"]:
            section_obj.extend(self._make_arg_items(info["parameters"],
                                                    ref_prefix))

        if info["returns"]:
            section_obj.extend(utils.parse_text(
                "**Returns**:\n%s" % info["returns"]))

        if info["schema"]:
            schema = process_jsonschema(info["schema"])
            if "type" in schema:
                if "parameters" in schema:
                    section_obj.extend(self._make_arg_items(
                        items=schema["parameters"],
                        ref_prefix=ref_prefix))
                elif "patternProperties" in schema:
                    section_obj.extend(self._make_arg_items(
                        items=schema["patternProperties"],
                        ref_prefix=ref_prefix,
                        description=["*Dictionary is expected. Keys should "
                                     "follow pattern(s) described bellow.*"]))
                elif "oneOf" in schema:
                    section_obj.append(utils.note(
                        "One of the following groups of "
                        "parameters should be provided."))
                    for i, oneOf in enumerate(schema["oneOf"], 1):
                        description = None
                        if oneOf.get("doc", None):
                            description = [oneOf["doc"]]
                        section_obj.extend(self._make_arg_items(
                            items=oneOf["parameters"],
                            ref_prefix=ref_prefix,
                            title="Option %s of parameters" % i,
                            description=description))
                else:
                    section_obj.extend(self._make_arg_items(
                        items=[schema], ref_prefix=ref_prefix))
            else:
                raise Exception("Failed to display provided schema: %s" %
                                info["schema"])

        if issubclass(plugin_cls, validation.ValidatablePluginMixin):
            validators = plugin_cls._meta_get("validators", default=[])
            platforms = [kwargs for name, args, kwargs in validators
                         if name == "required_platform"]
            if platforms:
                section_obj.append(
                    utils.paragraph("**Requires platform(s)**:"))
                section = ""
                for p in platforms:
                    section += "* %s" % p["platform"]
                    admin_msg = "credentials for admin user"
                    user_msg = ("regular users (temporary users can be created"
                                " via the 'users' context if admin user is "
                                "specified for the platform)")
                    if p.get("admin", False) and p.get("users", False):
                        section += " with %s and %s." % (admin_msg, user_msg)
                    elif p.get("admin", False):
                        section += " with %s." % admin_msg
                    elif p.get("users", False):
                        section += " with %s." % user_msg
                    section += "\n"

                section_obj.extend(utils.parse_text(section))

        filename = info["module"].replace(".", "/")
        if filename.startswith("rally/"):
            project = "rally"
        elif filename.startswith("rally_openstack/"):
            project = "rally-openstack"
        else:
            # WTF is it?!
            return None
        ref = ("https://github.com/openstack/%s/blob/master/%s.py"
               % (project, filename))
        section_obj.extend(utils.parse_text("**Module**:\n`%s`__\n\n__ %s"
                           % (info["module"], ref)))
        return section_obj

    def _make_plugin_base_section(self, base_cls, base_name=None):
        if base_name:
            title = ("%ss" % base_name if base_name[-1] != "y"
                     else "%sies" % base_name[:-1])
            subcategory_obj = utils.subcategory(title)
        else:
            subcategory_obj = []
        for p in sorted(base_cls.get_all(), key=lambda o: o.get_name()):
            # do not display hidden contexts
            if p._meta_get("hidden", False):
                continue
            subcategory_obj.append(self._make_plugin_section(p, base_name))

        return subcategory_obj

    @staticmethod
    def _parse_class_name(cls):
        name = ""
        for word in re.split(r"([A-Z][a-z]*)", cls.__name__):
            if word:
                if len(word) > 1 and name:
                    name += " "
                name += word
        return name

    @plugins.ensure_plugins_are_loaded
    def _get_all_plugins_bases(self):
        """Return grouped and sorted all plugins bases."""
        bases = []
        bases_names = []
        for p in discover.itersubclasses(plugin.Plugin):
            base_ref = getattr(p, "base_ref", None)
            if base_ref == p:
                name = self._parse_class_name(p)
                if name in bases_names:
                    raise Exception("Two base classes with same name '%s' are "
                                    "detected." % name)
                bases_names.append(name)
                category_of_base = "Common"
                for cname, cbases in CATEGORIES.items():
                    if name in cbases:
                        category_of_base = cname

                bases.append((category_of_base, name, p))
        return sorted(bases)

    def run(self):
        bases = self._get_all_plugins_bases()
        if "base_cls" in self.options:
            for _category_name, base_name, base_cls in bases:
                if base_name == self.options["base_cls"]:
                    return self._make_plugin_base_section(base_cls)
            raise Exception("Failed to generate plugins reference for '%s'"
                            " plugin base." % self.options["base_cls"])

        categories = {}

        for category_name, base_name, base_cls in bases:
            # FIXME(andreykurilin): do not ignore anything
            if base_name in IGNORED_BASES:
                continue
            if category_name not in categories:
                categories[category_name] = utils.category(category_name)
            category_of_base = categories[category_name]
            category_of_base.append(self._make_plugin_base_section(base_cls,
                                                                   base_name))
        return [content for _name, content in sorted(categories.items())]


def setup(app):
    plugins.load()
    app.add_directive("generate_plugin_reference", PluginsReferenceDirective)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.plugins.common.exporters.elastic import flatten
from tests.unit import test


class FlattenTestCase(test.TestCase):
    def test_transform(self):

        self.assertEqual(
            ["key1=value1", "key2=value2"],
            flatten.transform({"key1": "value1", "key2": "value2"}))

        self.assertEqual(
            ["key1=value1", "key2.foo.bar=1", "key2.xxx=yyy"],
            flatten.transform(
                {"key1": "value1", "key2": {"foo": {"bar": 1}, "xxx": "yyy"}}))

        self.assertEqual(
            ["foo[0]=xxx", "foo[1]=yyy", "foo[2].bar.zzz[0]=Hello",
             "foo[2].bar.zzz[1]=World!"],
            flatten.transform(
                {"foo": ["xxx", "yyy", {"bar": {"zzz": ["Hello", "World!"]}}]})
        )

    def test__join_keys(self):
        self.assertEqual("key", flatten._join_keys("key", ""))
        self.assertEqual("key.value", flatten._join_keys("key", "value"))
        self.assertEqual("[0].value", flatten._join_keys("[0]", "value"))
        self.assertEqual("key[0]", flatten._join_keys("key", "[0]"))
        self.assertEqual("[0][0]", flatten._join_keys("[0]", "[0]"))



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import subprocess

import ddt
import mock

from rally import consts
from rally.plugins.common.hook import sys_call
from rally.task import hook
from tests.unit import fakes
from tests.unit import test


@ddt.ddt
class SysCallHookTestCase(test.TestCase):

    @ddt.data(("ls", True), (50, False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = hook.HookAction.validate(
            "sys_call", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))

    @ddt.data(
        {"stdout": "foo output",
         "expected": {
             "additive": [],
             "complete": [{"chart_plugin": "TextArea",
                           "data": ["RetCode: 0", "StdOut: foo output",
                                    "StdErr: (empty)"],
                           "description": "Args: foo cmd",
                           "title": "System call"}]}},
        {"stdout": """{"additive": [],
                       "complete": [
                         {"chart_plugin": "Pie", "title": "Bar Pie",
                          "data": [["A", 4], ["B", 2]]}]}""",
         "expected": {
             "additive": [],
             "complete": [{"chart_plugin": "Pie", "data": [["A", 4], ["B", 2]],
                           "title": "Bar Pie"}]}})
    @ddt.unpack
    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    @mock.patch("rally.plugins.common.hook.sys_call.subprocess.Popen")
    def test_run(self, mock_popen, mock_timer, stdout, expected):
        popen_instance = mock_popen.return_value
        popen_instance.returncode = 0
        popen_instance.communicate.return_value = (stdout, "")
        hook = sys_call.SysCallHook(mock.Mock(), "foo cmd", {"iteration": 1})

        hook.run_sync()

        self.assertEqual(
            {"finished_at": fakes.FakeTimer().finish_timestamp(),
             "output": expected,
             "started_at": fakes.FakeTimer().timestamp(),
             "status": consts.HookStatus.SUCCESS,
             "triggered_by": {"iteration": 1}},
            hook.result())

        mock_popen.assert_called_once_with(["foo", "cmd"],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE,
                                           universal_newlines=True)

    @ddt.data(
        {"communicate_streams": ("foo out", "foo err"),
         "expected_error_details": "foo err",
         "expected_data_stderr": "StdErr: foo err"},
        {"communicate_streams": ("foo out", ""),
         "expected_error_details": "stdout: foo out",
         "expected_data_stderr": "StdErr: (empty)"})
    @ddt.unpack
    @mock.patch("rally.common.utils.Timer", side_effect=fakes.FakeTimer)
    @mock.patch("rally.plugins.common.hook.sys_call.subprocess.Popen")
    def test_run_error(self, mock_popen, mock_timer, communicate_streams,
                       expected_error_details, expected_data_stderr):
        popen_instance = mock_popen.return_value
        popen_instance.communicate.return_value = communicate_streams
        popen_instance.returncode = 1

        task = mock.MagicMock()
        sys_call_hook = sys_call.SysCallHook(task, "/bin/bash -c 'ls'",
                                             {"iteration": 1})

        sys_call_hook.run_sync()

        self.assertEqual(
            {"error": {"details": expected_error_details,
                       "etype": "n/a",
                       "msg": "Subprocess returned 1"},
             "finished_at": fakes.FakeTimer().finish_timestamp(),
             "output": {
                 "additive": [],
                 "complete": [{"chart_plugin": "TextArea",
                               "data": ["RetCode: 1",
                                        "StdOut: foo out",
                                        expected_data_stderr],
                               "description": "Args: /bin/bash -c 'ls'",
                               "title": "System call"}]},
             "started_at": fakes.FakeTimer().timestamp(),
             "status": "failed",
             "triggered_by": {"iteration": 1}}, sys_call_hook.result())

        mock_popen.assert_called_once_with(
            ["/bin/bash", "-c", "ls"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import datetime as dt

import mock

from rally.common import version as rally_version
from rally.plugins.common.exporters import json_exporter
from tests.unit.plugins.common.exporters import test_html
from tests.unit import test

PATH = "rally.plugins.common.exporters.json_exporter"


class JSONExporterTestCase(test.TestCase):

    def test__generate_tasks(self):
        tasks_results = test_html.get_tasks_results()
        reporter = json_exporter.JSONExporter(tasks_results, None)

        self.assertEqual([
            collections.OrderedDict([
                ("uuid", "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8"),
                ("title", "task"),
                ("description", "description"),
                ("status", "finished"),
                ("tags", []),
                ("env_uuid", "env-uuid"),
                ("env_name", "env-name"),
                ("created_at", "2017-06-04T05:14:44"),
                ("updated_at", "2017-06-04T05:15:14"),
                ("pass_sla", True),
                ("subtasks", [
                    collections.OrderedDict([
                        ("uuid", "subtask_uuid"),
                        ("title", "subtask"),
                        ("description", "description"),
                        ("status", "finished"),
                        ("created_at", "2017-06-04T05:14:44"),
                        ("updated_at", "2017-06-04T05:15:14"),
                        ("sla", {}),
                        ("workloads", [
                            collections.OrderedDict([
                                ("uuid", "uuid"),
                                ("description", "List all volumes."),
                                ("runner", {"runner_type": {}}),
                                ("hooks", []),
                                ("scenario", {
                                    "CinderVolumes.list_volumes": {}}),
                                ("min_duration", 0.0),
                                ("max_duration", 1.0),
                                ("start_time", 0),
                                ("load_duration", 2.03029203414917),
                                ("full_duration", 29.969523191452026),
                                ("statistics", {}),
                                ("data", {"raw": []}),
                                ("failed_iteration_count", 0),
                                ("total_iteration_count", 10),
                                ("created_at", "2017-06-04T05:14:44"),
                                ("updated_at", "2017-06-04T05:15:14"),
                                ("contexts", {}),
                                ("contexts_results", []),
                                ("position", 0),
                                ("pass_sla", True),
                                ("sla_results", {"sla": []}),
                                ("sla", {})
                            ])
                        ])
                    ])
                ])
            ])], reporter._generate_tasks())

    @mock.patch("%s.json.dumps" % PATH, return_value="json")
    @mock.patch("%s.dt" % PATH)
    def test_generate(self, mock_dt, mock_json_dumps):
        mock_dt.datetime.utcnow.return_value = dt.datetime.utcnow()
        tasks_results = test_html.get_tasks_results()

        # print
        reporter = json_exporter.JSONExporter(tasks_results, None)
        reporter._generate_tasks = mock.MagicMock()
        self.assertEqual({"print": "json"}, reporter.generate())
        results = {
            "info": {"rally_version": rally_version.version_string(),
                     "generated_at": mock_dt.datetime.strftime.return_value,
                     "format_version": "1.2"},
            "tasks": reporter._generate_tasks.return_value
        }
        mock_dt.datetime.strftime.assert_called_once_with(
            mock_dt.datetime.utcnow.return_value,
            json_exporter.TIMEFORMAT)
        reporter._generate_tasks.assert_called_once_with()
        mock_json_dumps.assert_called_once_with(results,
                                                sort_keys=False,
                                                indent=4)

        # export to file
        reporter = json_exporter.JSONExporter(tasks_results,
                                              output_destination="path")
        self.assertEqual({"files": {"path": "json"},
                          "open": "file://path"}, reporter.generate())
